#
# Copyright 2016 Universidad de Cantabria
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import os
import re
import glob
import time
import copy
import tarfile
import shutil
import logging
import fortran_namelist as fn
import re
from fortran_namelist       import coerce_value_list
from sqlalchemy             import and_, or_
from os.path                import ( exists, expandvars, 
                                     expanduser, isdir, 
                                     join, abspath )
from datetime               import datetime
from wrf4g                  import WRF4G_DIR, WRF4G_DEPLOYMENT_DIR
from wrf4g.config           import get_conf, save_json
from wrf4g.utils            import Enumerate, dict_compare 
from wrf4g.utils.archive    import extract
from wrf4g.utils.time       import ( datetime2datewrf, Calendar, 
                                     datewrf2datetime, timedelta_total_seconds )
from wrf4g.utils.file       import validate_name, edit_file
from wrf4g.utils.command    import exec_cmd
from wrf4g.utils.vcplib     import VCPURL
from wrf4g.utils.gridwaylib import GWJob


class Experiment(object):
    """ 
    Manage WRF4G experiments
    """
    dryrun = False
    
    def run(self, rerun = False, rea_pattern = False, rea_status = False, priority = 0 ):
        """
        Run the realizations of this experiment
        n_chunk is the number of chunks to run. 
        If there are not n_chunk, run every chunk since the last one finished 
        """
        #Check if the experiment have some realization to run
        #list of realizations of the experiment
        l_realizations  = self._filter_realizations( rea_pattern, rea_status )
        if not ( l_realizations ):
            logging.warn( 'There are not realizations to run.' )
        else:
            #if there are realizations to run
            for rea in l_realizations :
                logging.info( "---> Submitting Realization %s" % rea.name )
                #Run every realization
                rea.dryrun = self.dryrun
                if rerun :
                   first_chunk_run = 1
                else :
                   first_chunk_run = None
                rea.run( first_chunk_run = first_chunk_run, rerun = rerun, priority = priority )    

    def run_wps(self, rerun = False, rea_pattern = False, rea_status = False, priority = 0 ):
        """
        Run the realizations of this experiment
        n_chunk is the number of chunks to run. 
        If there are not n_chunk, run every chunk since the last one finished 
        """
        #Check if the experiment have some realization to run
        #list of realizations of the experiment
        l_realizations  = self._filter_realizations( rea_pattern, rea_status )
        if not ( l_realizations ):
            logging.warn( 'There are not realizations to run.' )
        else:
            #if there are realizations to run
            for rea in l_realizations :
                logging.info( "---> Submitting Realization %s" % rea.name )
                #Run every realization
                rea.dryrun = self.dryrun
                if rerun :
                   first_chunk_run = 1
                else :
                   first_chunk_run = None
                rea.run_wps( rerun = rerun, priority = priority )

    def edit(self):
        """
        Edit experiment.wrf4g file.
        """
        edit_file( join( self.home_directory, 'experiment.wrf4g' )  )

    def _copy_namelist_template( self, namelist_template ):
        """
        Copy the namelist from the template directory 
        """
        logging.info( "Preparing namelist %s version ... " % namelist_template )

        if not namelist_template.startswith("/"):
            if re.match("^\d+.\d+.\d+",namelist_template):
                namelist_template = join( WRF4G_DIR, 'etc', 'templates', 'namelist',
                                  'namelist.input-%s' % namelist_template )    
            else:
                raise Exception( "namelist version: %s is not an accepted value. It should point to a file or to a WRF version"  %namelist_template)
           
        namelist_input    = join( self.home_directory, 'namelist.input' )
        try :
            logging.debug( "Copying '%s' to '%s'" % ( namelist_template, namelist_input ) )
            shutil.copyfile( namelist_template, namelist_input )
        except :
            raise Exception( "There is not a namelist template for WRF '%s'"
                             "(File namelist.input does not exist)" % namelist_template )
        return namelist_input

    def check_db(self, name, start_date, end_date, cfg) :
        """ 
        Check if there is a realization with the same no reconfigurable field. 
        If there is not a realization with the same no reconfigurable fields => error
        If there is a realization with the same reconfigurable fields, update data
        """
        #Check if exists the realization
        try :
            rea = self.realization.filter_by( name = name ).one()
        except Exception :
            return None
        else :
            #Check if there is a realization with the same no reconfigurable fields
            if rea.cfg[ 'calendar' ] == cfg[ 'calendar' ] and rea.end_date != end_date :
                logging.debug( '\t\tUpdating realization on the database...' )
                rea.end_date = end_date
                rea.status   = Realization.Status.PREPARED
                return rea
            elif rea.end_date == end_date and rea.cfg[ 'calendar' ] == cfg[ 'calendar' ] :
                rea.cfg = cfg
                return rea
            else :                                 
                #if rea does not exist (no realization with the same no reconfigurable fields)
                raise Exception( "\t\tRealization with the same name and " 
                                 "no reconfigurable fields already exists" ) 

    def prepare(self, update = False, directory = './' ):
        """
        Prepare all realizations and chunks needed to submit a WRF4G experiment.
        """
        if update :
            directory = self.home_directory
        # Read experiment.wrf4g file
        self.cfg = get_conf( directory )
        if self.name != self.cfg[ 'default'][ 'name' ] :
            raise Exception( "ERROR: experiment.wrf4g file has a different experiment name." )
        
        # Update experiment variables
        self.name           = self.cfg[ 'default' ][ 'name' ]
        self.home_directory = abspath( directory ) 
   
        if not self.dryrun :
            exp_sub_dir = join( WRF4G_DIR, 'var', 'submission', self.name )
            if not isdir( exp_sub_dir ) :
                try:
                    logging.debug( "Creating '%s' directory" % exp_sub_dir )
                    os.makedirs( exp_sub_dir )
                except Exception :
                    raise Exception( "Couldn't be created '%s' directory" % exp_sub_dir )
          
        # Cycle to create a realization per combination
        for section in sorted( list( self.cfg.keys() ) ) :
            if section.startswith( "ensemble" ) :
                try :
                    # Copy the namelist from the template directory 
                    namelist_input = self._copy_namelist_template( self.cfg[ section ][ 'namelist_template' ] )
                    self.update_namelist( namelist_input, section )
                    self.cycle_time( namelist_input, section )
                except KeyError as err :
                    logging.error( "%s is a mandatory variable."
                                   " Please add this variable to experiment.wrf4g file" % str(err) )

        if not self.dryrun :
            # Copy configure files before submission
            self._copy_experiment_files( exp_sub_dir  )
            # Create software bundles to use on the WN
            self._create_wrf4g_bundles( exp_sub_dir  )
    
    def get_status(self, rea_pattern = False, rea_status = False, showchunks= False ):
        """ 
        Show information about realizations of the experiment for example:
        
        Realization Status   Chunks     Comp.Res      Run.Sta     JID ext      %
        testc       Finished    3/3     mycomputer   Finished       0   0 100.00
        
        * Realization: Realization name. It is taken from the field experiment_name in experiment.wrf4g.
        * Status: It can be take the following values: Prepared, Submitted, Running, Failed and Done).
        * Chunks [Chunk running/Total Chunks]: A realization is split into chunks. Each chunk is sent as a job.
        * Computer resource: Resource (cluster) where the job is running.
        * Run.Sta: Job status in the WN (Downloading data, running ungrib, real, wrf, ...)
        * JID: Job identifier.
        * ext: Exit Code. If exit code is different from 0, there has been an error. 
        * % : percentage of simulation finished.
        """
        #list of realization of the experiment
        l_realizations = self._filter_realizations( rea_pattern, rea_status )
        #Header of the information
        if not ( l_realizations ):
            raise Exception ( 'There are not realizations to check.' )
        if showchunks:
            Realization.status_chunk_header()
        else:
            Realization.status_header()

        for rea in l_realizations.order_by( Realization.name ) :
            #Print information of each realization
            rea.get_status(showchunks)
    
    def cancel(self, rea_pattern = False, rea_status = False, hard = False ):
        """
        Delete jobs which status is running or submitted 
        """
        #list of realization of the experiment
        l_realizations  = self._filter_realizations( rea_pattern, rea_status )
        if not ( l_realizations ):
            logging.info( 'There are not realizations to cancel.' )
        else :
            logging.info( 'Canceling Experiment %s' % self.name )
            for rea in l_realizations :
                rea.dryrun = self.dryrun
                rea.cancel( hard )
 
    def release(self):
        """
        Check created job to be release
        """
        l_realizations  = self._filter_realizations( False, Realization.Status.SUBMITTED )
        for rea in l_realizations :
            rea.release( ) 

    def release_wps(self):
        """
        Check created job to be release
        """
        l_realizations  = self._filter_realizations( False, Realization.wps_only_Status.WPS_SUBMITTED, wps_only=True )
        for rea in l_realizations :
            rea.release_wps( ) 

    def statistics(self, rea_pattern = False ):
        """
        Get statistics from the database 
        """
        #list of realization of the experiment
        l_realizations  = self._filter_realizations( rea_pattern, False )
        logging.info( "Job ID;Realization Name;Chunk ID;Resource Name;Execution Time (s);"
                      "Waiting Time (s);REAL Execution Time (s);WRF Execution Time (s)" )
        for rea in l_realizations :
            rea.statistics( )

    def set_priority(self, rea_pattern = False, priority = 0 ):
        """
        Setting priority to jobs 
        """
        #list of realization of the experiment
        l_realizations  = self._filter_realizations( rea_pattern, False )
        if not ( l_realizations ):
            logging.info( 'There are not realizations to set priority.' )
        else :
            logging.info( 'Setting priority Experiment %s' % self.name )
            for rea in l_realizations :
                rea.dryrun = self.dryrun
                rea.set_priority( priority )

    def delete(self):
        """
        Delete the experiment, its realizations and chunks 
        """
        # Delete the local submission directory
        local_exp_dir = join( WRF4G_DIR , 'var' , 'submission' , self.name )
        logging.debug( "Deleting '%s' directory" % local_exp_dir )
        if not self.dryrun :
            if exists( local_exp_dir ) : 
                shutil.rmtree( local_exp_dir )

    @staticmethod 
    def create_files(name, template, force, directory):
        """
        Create the files needed to establish a WRF4G experiment
        """
        validate_name( name )
        template_dir = join( WRF4G_DIR , 'etc' , 'templates' , 'experiments',  template )
        if not exists(template_dir):
            raise Exception( "'%s' template does not exist" % template )
        exp_dir = expandvars( expanduser( os.getcwd() if directory == './' else directory ) )
        if not exists( exp_dir ):
            raise Exception("'%s' does not exist" % exp_dir )
        exp_dir_config = join( exp_dir, name )
        if exists( exp_dir_config ) and not force :
            raise Exception("'%s' already exists" % exp_dir_config )
        elif exists( exp_dir_config ) and force :
            shutil.rmtree( exp_dir_config )
        logging.debug( "Creating '%s' directory" % exp_dir_config )
        shutil.copytree( join( WRF4G_DIR , 'etc' , 'templates' , 'experiments',  template ),
                         exp_dir_config )
        dest_path = join( exp_dir_config , 'experiment.wrf4g' )
        with open( dest_path , 'r') as f :
            data = ''.join( f.readlines( ) )
        data_updated = data % {
                               'WRF4G_EXPERIMENT_HOME' : exp_dir_config ,
                               'WRF4G_DEPLOYMENT_DIR'  : WRF4G_DEPLOYMENT_DIR ,
                               'WRF4G_DIR'             : WRF4G_DIR ,
                               'exp_name'              : name ,
                               }
        with open(dest_path, 'w') as f :
            f.writelines( data_updated )
 
    def update_namelist( self, namelist_input, section ):
        """
        Update namelist values
        """
        nmli = fn.WrfNamelist( namelist_input )
        logging.debug( "Updating parameter 'max_dom' in the namelist" )
        nmli.setValue( "max_dom", int( self.cfg[ section ][ 'max_dom' ] ) )
        max_dom = single = False
        
        # If namelist_values is empty, do not modify namelist variables
        if (len(self.cfg[ section ][ 'namelist_values']) != 0):
            for mnl_variable, mnl_values in self.cfg[ section ][ 'namelist_values' ].items() :
                # Update the namelist per each combination
                logging.debug( "Updating parameter '%s' in the namelist" % mnl_variable )
                # Modify the namelist with the parameters available in the namelist description
                if mnl_variable.startswith( "max_dom:" ) :
                    mnl_variable = mnl_variable[ 8: ]
                    max_dom      = True
                elif mnl_variable.startswith( "single:" ) :
                    mnl_variable = mnl_variable[ 7: ]
                    single       = True
                if '.' in mnl_variable :
                    nml_section, val = mnl_variable.split( '.' )
                else :
                    nml_section, val = "",  mnl_variable
                if max_dom and not val in nmli.MAX_DOM_VARIABLES :
                    nmli.MAX_DOM_VARIABLES.extend( val  )
                if single and val in nmli.MAX_DOM_VARIABLES :
                    nmli.MAX_DOM_VARIABLES.remove( val  )
                try :
                    nmli.setValue( val, coerce_value_list( mnl_values.strip( ',' ).split( ',' ) ), nml_section )
                except IndexError:
                    raise Exception( "'%s' does not have values for all namelist combinations." % mnl_variable )
                except Exception as err:
                    raise Exception( err )
        nmli.trimMaxDom()
        nmli.extendMaxDomVariables()
        if nmli.wrfCheck() :
            raise Exception( "Please review 'namelist_values' variable." )
        if not self.dryrun :
            nmli.overWriteNamelist()

    def _update_member( self, realization_start_date ) :
        """
        If the member is not indicated the initialized month will be the 
        month of the realization
        """
        runtime = []
        for elem in range( self.cfg[ section ] [ 'preprocessor_optargs' ] [ 'member' ].count( ',' ) ) :
            runtime.append( str( realization_start_date.month ) )
        self.cfg[ section ] [ 'preprocessor_optargs' ] [ 'runtime' ] = ','.join( runtime )

    def cycle_time( self, namelist_input, section ) :
        """
        Clycle realization time
        """
        realization_name = self.name + '-' + section.split( "/" )[ 1 ]
        for ( start_date, end_date, simult_interval,
              simult_length, chunk_size, restart_interval ) in self.cfg[ section ][ 'date_time' ] :
            # Update restart_interval in the namelist 
            logging.debug( "Updating parameter 'restart_interval' in the namelist" )
            nmli = fn.WrfNamelist( namelist_input )
            nmli.setValue( "restart_interval", restart_interval )
            if not self.dryrun :
                nmli.overWriteNamelist()
            # Define which calendar is going to be used
            exp_calendar   = Calendar( self.cfg[ section ][ 'calendar' ] )
            rea_start_date = start_date
            while rea_start_date < end_date :
                rea_end_date = exp_calendar.add( rea_start_date, simult_length )
                if rea_end_date > end_date :
                    rea_end_date = end_date
                rea_name = "%s-%s" % ( realization_name, rea_start_date.strftime( "%Y%m%dT%H%M%S" ) )
                logging.info( "---> Realization %s: start date %s end date %s" % (
                               rea_name, rea_start_date, rea_end_date ) )
                # Check realization on the database
                rea = self.check_db( name = rea_name, start_date = rea_start_date, end_date = rea_end_date,
                                     cfg = self.cfg[ section ] )
                # Create chunks only if end date has been modified
                if rea:
                    rea.cycle_chunks()
                elif not rea:
                    # If there is not runtime we have to add the start month of the realization
                    if ( 'preprocessor_optargs' in  self.cfg[ section ] ) and \
                       ( 'member' in self.cfg[ section ] [ 'preprocessor_optargs' ] ) and \
                       (  not 'runtime' in self.cfg[ section ] [ 'preprocessor_optargs' ] ) :
                        self._update_member( rea_start_date )
                    # Create a realization 
                    rea = Realization( )
                    rea.name             = rea_name
                    rea.start_date       = rea_start_date
                    rea.end_date         = rea_end_date
                    rea.chunk_size       = chunk_size
                    rea.current_date     = rea_start_date
                    rea.status           = Realization.Status.PREPARED
                    rea.current_chunk    = 1
                    rea.cfg              = self.cfg[ section ]
                    # Add realization to the experiment 
                    self.realization.append( rea )
                    # Create chunk for the realization
                    rea.cycle_chunks()
                # Check storage
                if not self.dryrun :
                    # Default section will be the current section
                    realization_cfg = dict()
                    realization_cfg[ 'ensemble/default' ] = copy.deepcopy( self.cfg[ section ] )
                    for key, val in self.cfg[ section ].items() :
                        if key.startswith( 'resource' ) :
                            realization_cfg[ key ] = copy.deepcopy( val )
                    save_json( realization_cfg, self.home_directory, "realization.json" )
                    rea._prepare_sub_files()
                rea_start_date = exp_calendar.add( rea_start_date, simult_interval )

    def _copy_experiment_files(self, exp_sub_dir ):
        """
        Copy configure files before submission.
        """    
        for file in [ join( WRF4G_DIR, "etc", "db.conf" ),
                      join( self.home_directory, "experiment.wrf4g" ) ] :
            if not exists( expandvars( file ) ) :
                raise Exception( "'%s' is not available" % file )
            else :
                shutil.copy( expandvars( file ) , exp_sub_dir )

    def _create_wrf4g_bundles(self, exp_sub_dir):
        """
        Create bundles with the necessary software to run WRF on worker nodes.
        """
        # WRF4G bundle
        logging.debug( "Create a WRF4G software bundle to use on the worker node..." )
        wrf4g_package = join ( exp_sub_dir , "WRF4G.tar.gz" )
        if exists( wrf4g_package  ):
            logging.debug( "Removing '%s' package" % wrf4g_package )
            os.remove( wrf4g_package )
        current_path = os.getcwd()
        try :
            tar = tarfile.open( wrf4g_package, "w:gz" )
            # Add wn/bin
            tar.add('%s/data/wn/bin' % (WRF4G_DEPLOYMENT_DIR),arcname='bin')
            # Add python packages to lib/python
            for package in [ 'sqlalchemy','dateutil','wrf4g','fortran_namelist','drm4g']:
                ipackage = __import__(package)
                tar.add(os.path.dirname(ipackage.__file__),arcname='lib/python/%s' % (package) )
            for module in ['six']:
                imodule = __import__(module)
                tar.add('%s/%s.py' %(os.path.dirname(imodule.__file__),module), arcname='lib/python/%s' % ('%s.py' %(module)))
        except Exception as err:
            logging.warn( err )
        finally :
            tar.close()
        # wrf4g_files bundle
        wrf4g_files_dir = join( self.home_directory, 'wrf4g_files' )
        if isdir( wrf4g_files_dir ):
            logging.debug( "Create a wrf4g_files.tar.gz bundle to use on the worker node..." )
            wrf4g_files_package = join ( exp_sub_dir , "wrf4g_files.tar.gz" )
            if exists( wrf4g_files_package ):
                logging.debug( "Removing '%s' package" % wrf4g_files_package )
                os.remove( wrf4g_files_package )
            tar = tarfile.open( wrf4g_files_package , "w:gz" )
            os.chdir( wrf4g_files_dir )
            for elem in os.listdir('.') :
                tar.add( elem )
            tar.close()
        os.chdir( current_path )

    def _filter_realizations(self, pattern, status, wps_only=False):
        """
        Filter realizations from the experiment
        """
        l_realizations = self.realization
        if pattern :
            l_realizations = l_realizations.\
                             filter( Realization.name.like( pattern.replace( '*', '%' )\
                                                                   .replace( '?', '_' ) ) ) 
        if status and not wps_only:
            l_realizations = l_realizations.\
                             filter_by( status = status )
        elif status and wps_only:
            l_realizations = l_realizations.\
                             filter_by( wps_only_status = status )
        return l_realizations
    
class Realization( object ):
    """
    A class to mange WRF4G realizations
    """
    dryrun = False

    # Status = Enumerate( 'PREPARED', 'SUBMITTED','WPS_SUBMITTED', 'RUNNING',
    #                     'PENDING', 'FAILED', 'FINISHED','WPS_FINISHED' )

    Status = Enumerate( 'PREPARED', 'SUBMITTED', 'RUNNING',
                         'PENDING', 'FAILED', 'FINISHED')
    
    wps_only_Status = Enumerate( 'PREPARED', 'WPS_SUBMITTED', 'RUNNING',
                         'PENDING', 'FAILED', 'WPS_FINISHED')
        
    def run(self, first_chunk_run = None , last_chunk_run = None, rerun = False, priority = 0 ):
        """ 
        Run n_chunk of the realization.
        If n_chunk=0 run every chunk of the realization which haven't finished yet
        else run (n_chunk) chunks since the last one finished
        """
        first_chunk_run = int( first_chunk_run ) if first_chunk_run else None 
        last_chunk_run  = int( last_chunk_run  ) if last_chunk_run  else None
        #Check the status of the realization
        if self.status == Realization.Status.FINISHED and not rerun:
            logging.warn( "\tRealization '%s' already finished." % self.name )
        elif ( self.status == Realization.Status.SUBMITTED or 
               self.status == Realization.Status.RUNNING ) and not rerun :
            logging.warn( "\tRealization '%s' has been submitted." % self.name )
        elif first_chunk_run and first_chunk_run < 0 :
            logging.error( "\tERROR: The first chunk to run is '%d'." % first_chunk_run ) 
        elif last_chunk_run and last_chunk_run  < 0 :
            logging.error( "\tERROR: The last chunk to run is '%d'." % last_chunk_run )
        elif ( last_chunk_run and first_chunk_run ) and last_chunk_run < first_chunk_run :
            logging.error( "\tERROR: The last chunk to run is greater than the fist one." )
        elif last_chunk_run and last_chunk_run > self.nchunks :
            logging.error( "\tERROR: The last chunk does not exist." )
        elif first_chunk_run and first_chunk_run > self.nchunks :
            logging.error( "\tERROR: The first chunk does not exist." )
        else :
            # search first chunk to run
            if rerun and first_chunk_run :
                ch                 = self.chunk.filter( Chunk.chunk_id == first_chunk_run ).one()
                self.restart       = ch.start_date
                self.current_date  = ch.start_date
                self.current_chunk = first_chunk_run
            elif rerun and not first_chunk_run :
                self.restart       = None
                first_chunk_run    = self.current_chunk = 1
            else :
                #search first chunk to run
                if not self.restart : # run every chunks of the realization
                    first_chunk_run = 1
                else:
                    #search chunk with end_date>restart and start_date<restart
                    try:
                        first_chunk  = self.chunk.filter( and_( Chunk.start_date <= self.restart, 
                                                                Chunk.end_date   >= self.restart ) 
                                                         ).all()[ -1 ]
                    except : 
                        raise Exception( 'There are not chunks to run.' )
                    else:
                        if first_chunk_run and first_chunk.chunk_id != first_chunk_run :
                            raise Exception( 'Use the option --rerun.' )
                        else : 
                            first_chunk_run = self.current_chunk = first_chunk.chunk_id
            #search last chunk to run
            if not last_chunk_run :
                #run every chunk
                #Search last chunk of the realization
                last_chunk_run = self.nchunks
            else:
                #search last chunk
                last_chunk_run = last_chunk_run
            #Search chunks to run
            l_chunks = self.chunk.filter( and_( Chunk.chunk_id >= first_chunk_run, 
                                                Chunk.chunk_id <= last_chunk_run )
                                        ).all()
            #run chunks
            for index, chunk in enumerate( l_chunks ) :
                #print data of chunks
                logging.info( '\t---> Submitting Chunk %d %s %s' % ( chunk.chunk_id, 
                                                              datetime2datewrf(chunk.start_date), 
                                                              datetime2datewrf(chunk.end_date) ) )
                if not self.dryrun :
                    chunk.run( index, rerun, priority,mode='wrf' )
            if not self.dryrun :
                # Update realizaiton status
                self.status = Realization.Status.SUBMITTED
            
    
    def run_wps(self, first_chunk_run = None , last_chunk_run = None, rerun = False, priority = 0 ):
        """ 
        Run n_chunk of the realization.
        If n_chunk=0 run every chunk of the realization which haven't finished yet
        else run (n_chunk) chunks since the last one finished
        """
        first_chunk_run = int( first_chunk_run ) if first_chunk_run else None 
        last_chunk_run  = int( last_chunk_run  ) if last_chunk_run  else None
        #Check the status of the realization
        #if self.status == Realization.Status.WPS_FINISHED and not rerun:
        #    logging.warn( "\tWPS step of Realization '%s' already finished." % self.name )
        if self.status == Realization.Status.FINISHED and not rerun:
            logging.warn( "\tRealization '%s' already finished." % self.name )
        elif ( self.status == Realization.Status.SUBMITTED or 
               self.status == Realization.Status.RUNNING ) and not rerun :
            logging.warn( "\tRealization '%s' has been submitted." % self.name )
        elif first_chunk_run and first_chunk_run < 0 :
            logging.error( "\tERROR: The first chunk to run is '%d'." % first_chunk_run ) 
        elif last_chunk_run and last_chunk_run  < 0 :
            logging.error( "\tERROR: The last chunk to run is '%d'." % last_chunk_run )
        elif ( last_chunk_run and first_chunk_run ) and last_chunk_run < first_chunk_run :
            logging.error( "\tERROR: The last chunk to run is greater than the fist one." )
        elif last_chunk_run and last_chunk_run > self.nchunks :
            logging.error( "\tERROR: The last chunk does not exist." )
        elif first_chunk_run and first_chunk_run > self.nchunks :
            logging.error( "\tERROR: The first chunk does not exist." )
        else :
            # search first chunk to run
            if not first_chunk_run :
                first_chunk_run = 1
            
            #search last chunk to run
            if not last_chunk_run :
                #run every chunk
                #Search last chunk of the realization
                last_chunk_run = self.nchunks

            #Search chunks to run
            l_chunks = self.chunk.filter( and_( Chunk.chunk_id >= first_chunk_run, 
                                                Chunk.chunk_id <= last_chunk_run )
                                        ).all()
            #run chunks
            for index, chunk in enumerate( l_chunks ) :
                #print data of chunks
                logging.info( '\t---> Submitting Chunk %d %s %s' % ( chunk.chunk_id, 
                                                              datetime2datewrf(chunk.start_date), 
                                                              datetime2datewrf(chunk.end_date) ) )
                if not self.dryrun :
                    chunk.run( index, rerun, priority, mode='wps-only' )
            if not self.dryrun :
                # Update realizaiton status
                self.wps_only_status = Realization.wps_only_Status.WPS_SUBMITTED
    
    def _prepare_sub_files(self):
        """
        Prepare the files needed to submit the realization. 
        """
        rea_submission_path = join( WRF4G_DIR, 'var', 'submission', self.experiment.name, self.name )
        if not isdir( rea_submission_path ) :
            try :
                os.makedirs( rea_submission_path )
            except Exception :
                raise Exception( "Couldn't be created '%s' directory" % rea_submission_path )
        for file_name in [ join( self.experiment.home_directory, "namelist.input" ),
                           join( self.experiment.home_directory, "realization.json" ) ] :
            if not exists( file_name ) :
                raise Exception( "'%s' is not available" % file_name )
            else :
                shutil.copy( file_name , rea_submission_path )

    def check_db(self, rea_id, chunk_start_date, chunk_end_date, chunk_id ):
        """ 
        Check if there is a chunk with the same no reconfigurable field. 
        If there is not a chunk with the same no reconfigurable fields => error,.
        If there is a chunk with the same reconfigurable fields, update data.
        """
        #Check if there is a chunk with the same fields
        try:
            ch = self.chunk.filter( Chunk.rea_id     == rea_id,
                                    Chunk.chunk_id   == chunk_id,
                                    Chunk.start_date == chunk_start_date,
                                    Chunk.end_date   == chunk_end_date,
                                   ).one()
        except Exception :
            #There will be an exception for the last chunk of a realization.
            #This chunk will be able to modify 
            if self.nchunks == chunk_id :
                try :
                    ch2 = self.chunk.filter( Chunk.rea_id     == rea_id,
                                             Chunk.chunk_id   == chunk_id,
                                             Chunk.start_date == chunk_start_date,
                                             Chunk.end_date   != chunk_end_date,
                                           ).one()
                except Exception :
                    return None
                else :
                    logging.debug( '\t\t\tUpdating chunk on the database...' )
                    ch2.end_date = chunk_end_date
                    return ch2
            else :
                return None
        else :
            #if ch exists 
            return ch

    def cycle_chunks(self):
        """
        Create chunks the needed for a realization 
        """
        # Define which calendar is going to be used
        exp_calendar = Calendar( self.cfg[ 'calendar' ] )
        chunk_id = 1
        chunk_start_date = self.start_date
        while chunk_start_date < self.end_date :
            chunk_end_date = exp_calendar.add( chunk_start_date, self.chunk_size )
            if chunk_end_date > self.end_date :
                chunk_end_date = self.end_date
            # Check chunk on the database
            ch = self.check_db( rea_id           = self.id, 
                                chunk_start_date = chunk_start_date, 
                                chunk_end_date   = chunk_end_date,
                                chunk_id         = chunk_id
                                )
            if not ch :
                logging.info( "\t\t---> Chunk %d %s %s" %( chunk_id,
                                                       datetime2datewrf(chunk_start_date),
                                                       datetime2datewrf(chunk_end_date) ) )
                # Create Chunk
                ch = Chunk( )
                ch.rea_id     = self.id
                ch.start_date = chunk_start_date
                ch.end_date   = chunk_end_date
                ch.wps        = 0
                ch.chunk_id   = chunk_id
                ch.status     = Chunk.Status.PREPARED
                # Add realization to the experiment 
                self.chunk.append( ch )
            chunk_start_date = chunk_end_date 
            chunk_id         = chunk_id + 1
        # Set the number of chunks of a relaization    
        self.nchunks = chunk_id - 1

    @staticmethod 
    def status_header(): 
        logging.info( '%-60s %-10s %-10s %-16s %-10s %6s %-3s %6s'% (
                        'REALIZATION','STATUS','CHUNKS','RESOURCE','RUN STATUS',
                        'JID', 'EXT','%' ) )
    @staticmethod 
    def status_chunk_header(): 
        logging.info( '%-60s %-10s %-20s %-20s %-10s %16s'% (
                        'REALIZATION','CHUNK_ID','START','END','WPS','STATUS' ) )
 
    def get_status(self,showchunks=0):
        """ 
        Show information about the realization for example:
            Realization Status    Chunks     Comp.Res       Run.Sta     JID   ext      %
            testc       Finished     3/3   mycomputer      Finished       0     0 100.00
            
        * Realization: Realization name. It is taken from the field experiment_name in experiment.wrf4g.
        * Status: It can be take the following values: Prepared, Submitted, Running, Failed and Done).
        * Chunks [Chunk running/Total Chunks]: A realization is split into chunks. Each chunk is sent as a job.
        * Computer resource: Resource (cluster) where the job is running.
        * Run.Sta: Job status in the WN (Downloading data, running ungrib, real, wrf, ...)
        * JID: Job identifier
        * ext: Exit Code. If exit code is different from 0, there has been an error.
        * % : percentage of simulation finished.
        """                
        #Select parameters:job_status,rea_status,resource,exitcode,nchunks
        #Last job of current chunk

        if not showchunks:
            resource, exitcode, gw_job = '-', '-', '-'
            status = Realization.Status.PREPARED
            if self.status != Realization.Status.PREPARED :
                ch = self.chunk.filter_by( chunk_id = self.current_chunk ).one()
                try :
                    last_job = ch.job[ -1 ]
                except :
                    pass
                else :
                    resource = last_job.resource
                    exitcode = last_job.exitcode
                    status   = last_job.status
                    gw_job   = last_job.gw_job
            chunk_distribution = '%d/%d' % ( 0 if not self.current_chunk else self.current_chunk, self.nchunks )
            #Format chunks run / chunks total
            runt   = int( self.current_date.strftime("%s") ) - int( self.start_date.strftime("%s") ) 
            totalt = int( self.end_date.strftime("%s") )     - int( self.start_date.strftime("%s") )
            #Percentage
            per = runt * 100.0 / totalt
            #Print output
            logging.info( "%-60s %-10.10s %-10.10s %-16.16s %-10.10s %6.6s %-3.3s %6.2f" % (
                        self.name, self.status, chunk_distribution, resource, status, 
                        gw_job, exitcode, per ) )

        else:
            for ch in self.chunk:
                logging.info( "%-60s %-10s %-20s %-20s %-10s %16s"% (
                    self.name, ch.chunk_id, ch.start_date, ch.end_date, ch.wps, ch.status ) )


    def information( self ) :
        """
        Get information about all realization values.
        """
        logging.info( "Start date: %s" % self.start_date )
        logging.info( "End date: %s" % self.end_date  )
        logging.info( "Current date: %s" % self.current_date )
        logging.info( "Chunk size: %s" % self.chunk_size )
        logging.info( "Number of chunks: %s" % self.nchunks )
        logging.info( "Status: %s" % self.status )
        logging.info( "Configuration: " )
        for key, val in self.cfg.items() :
           if 'date_time' in key : 
               continue
           elif 'app' in key :
               logging.info( "\t%s: %s" % ( key, val.replace( '\n', '\n\t\t' ) ) )
           elif type( val ) == dict :
               logging.info( "\t%s" % key )
               for key2, val2 in val.items() :
                   logging.info( "\t\t%s: %s" % ( key2, val2 ) )
           elif type( val ) == list or type( val ) == tuple :
               logging.info( "\t%s: %s" % ( key, ' '.join( val ) ) )
           else :
               if 'app' in key : val
               logging.info( "\t%s: %s" % ( key, val ) )
  
    def get_log( self, chunk_id , directory ) :
        """
        Search and unpack log files. 
        """
        tar_path      = join( WRF4G_DIR, 'var', 'submission', self.experiment.name, self.name )
        all_tar_files = glob.glob( join( tar_path, 'log_%s_*.tar.gz' % chunk_id ) ) 
        if not all_tar_files :
            raise Exception( 'There is not a log available for this chunk.' )
        for tar_file in all_tar_files :
            logging.info( "Unpacking %s file in the %s directory" % ( tar_file, directory ) )
            extract( tar_file, expandvars( expanduser( directory ) ) )

    def cancel(self, hard = False ):
        """
        Delete chunks which status is running or submitted.
        """
        l_chunks = self.chunk.filter( or_( Chunk.status == Chunk.Status.SUBMITTED,
                                           Chunk.status == Chunk.Status.PENDING,
                                           Chunk.status == Chunk.Status.FAILED,
                                           Chunk.status == Chunk.Status.RUNNING ) 
                                    ).all()
        logging.info( '---> Canceling Realization %s' % self.name )
        if not ( l_chunks ):
            logging.info( '\tThere are not chunks to cancel.' )
        else :
            for chunk in l_chunks :
                chunk.dryrun = self.dryrun
                chunk.cancel( hard )

    def release(self):
        """
        Check created job to be release
        """
        current_ch = self.chunk.filter( and_( Chunk.chunk_id == self.current_chunk,
                                              Chunk.status   == Chunk.Status.SUBMITTED )
                                      ).first()
        if current_ch :
            l_jobs     = current_ch.job.filter( Job.status == Job.Status.SUBMITTED )
            try :
                job = l_jobs[ -1 ]
            except :
                pass
            else :
                logging.debug( "Releasing job %s" % job.gw_job )
                GWJob().release( job.gw_job  )
                job.set_status( Job.Status.RELEASED )
    
    def release_wps(self):
        """
        Check created job to be release
        """
        current_ch = self.chunk.filter(Chunk.status==Chunk.Status.WPS_SUBMITTED).all()
        for chunk in current_ch:
            l_jobs     = chunk.job.filter( Job.status == Job.Status.SUBMITTED )
            try :
                job = l_jobs[ -1 ]
            except :
                pass
            else :
                logging.debug( "Releasing job %s" % job.gw_job )
                GWJob().release( job.gw_job  )
                job.set_status( Job.Status.RELEASED )


    def statistics(self):
        """
        Get statistics from jobs.
        """
        for chunk in self.chunk.all() :
            if chunk.job.all() [ -1 ].status == Job.Status.FINISHED :
                finished_job   = chunk.job.filter_by( status = Job.Status.FINISHED ) [ -1 ]
                date_SUBMITTED = finished_job.events.filter_by( job_status = Job.Status.SUBMITTED ).one().timestamp
                date_RUNNING   = finished_job.events.filter_by( job_status = Job.Status.RUNNING ).one().timestamp
                date_REAL      = finished_job.events.filter_by( job_status = Job.Status.REAL ).one().timestamp
                date_WRF       = finished_job.events.filter_by( job_status = Job.Status.WRF ).one().timestamp
                date_FINISHED  = finished_job.events.filter_by( job_status = Job.Status.FINISHED ).one().timestamp
                logging.info( "%d;%d;%s;%s;%d;%d;%d;%d" % ( finished_job.gw_job, chunk.chunk_id, self.name, finished_job.resource,
                                                        timedelta_total_seconds( date_FINISHED - date_RUNNING ), 
                                                        timedelta_total_seconds( date_RUNNING  - date_SUBMITTED ),
                                                        timedelta_total_seconds( date_WRF      - date_REAL ),
                                                        timedelta_total_seconds( date_FINISHED - date_WRF ) ) )

    def set_priority(self, priority = 0 ):
        """
        Setting priority to chunks which status is submitted.
        """
        if priority < 0 or priority > 20 :
            raise Exception( "'%d' priority is out of the range [0, 20]" % priority )      
        l_chunks = self.chunk.filter_by( status = Chunk.Status.SUBMITTED ).all()
        logging.info( '---> Setting priority Realization %s' % self.name )
        if not ( l_chunks ):
            logging.info( '\tThere are not chunks to set priority.' )
        else :
            for chunk in l_chunks :
                chunk.dryrun = self.dryrun
                chunk.set_priority( priority )
   
    def get_restart(self):
        """
        Get restart date.
        """
        logging.info( datetime2datewrf( self.restart ) )
 
    def set_restart(self, restart_date ):
        """
        Setting restart date.
        """
        try :
            datetime_restart_date = datewrf2datetime( restart_date ) 
        except :
            raise Exception( "ERROR: restart date is malformed" )
        else :
            logging.info( '---> Setting restart date %s' % datetime2datewrf( datetime_restart_date ) )
            self.restart = datetime_restart_date
 
class Chunk( object ):
    """ 
    A class to manage WRF4G chunks
    """
    dryrun = False 

    Status = Enumerate( 'PREPARED', 'SUBMITTED','WPS_SUBMITTED', 'RUNNING',
                        'PENDING', 'FAILED', 'FINISHED','WPS_FINISHED' )
 
    #METHODS
    def run (self, index, rerun = False, priority = 0, mode='wrf' ):
        """ 
        Run a chunk is run a drm4g job
        """
        #Send a gridway's job and save data in table Job
        gw_job        = GWJob()
        # create template
        rea_name      = self.realization.name
        exp_name      = self.realization.experiment.name
        exp_path      = join( WRF4G_DIR, 'var', 'submission', exp_name )
        rea_path      = join( exp_path, rea_name )
        wrf4g_package = join( exp_path, "WRF4G.tar.gz" )
        if not exists(  wrf4g_package ) : 
            raise Exception( "'%s' file does not exist" % wrf4g_package )
        # files to add for the inputsandbox 
        inputsandbox  = "file://%s,"                  % wrf4g_package
        inputsandbox += "file://%s/db.conf,"          % exp_path
        inputsandbox += "file://%s/experiment.wrf4g," % exp_path
        inputsandbox += "file://%s/realization.json," % rea_path  
        inputsandbox += "file://%s/namelist.input"    % rea_path  
        # Add input file if it is exist
        input_files = join( exp_path, 'wrf4g_files.tar.gz' )
        if exists( input_files ):
            inputsandbox += ",file://%s" % ( input_files )
        # files to add for the outputsandbox
        outputsandbox = "log_%d_${JOB_ID}.tar.gz, events.pkl" % self.chunk_id
        arguments = '%s %s %d %s %s %d %s' % ( exp_name, rea_name, self.chunk_id,
                                            datetime2datewrf( self.start_date ),
                                            datetime2datewrf( self.end_date ),
                                            1 if rerun else 0,
                                            mode
                                             )
        # Create the job template
        file_template = gw_job.create_template( name          = rea_name,
                                                directory     = rea_path,
                                                arguments     = arguments,
                                                np            = int( self.realization.cfg.get( 'np', '1' ) ),
                                                req           = self.realization.cfg.get( 'requirements', '' ),
                                                environ       = self.realization.cfg.get( 'environment', '' ) ,
                                                inputsandbox  = inputsandbox,
                                                outputsandbox = outputsandbox )
        # Submit the template
        job = Job()  #create an object "job"
        time.sleep( 0.1 )
        # if the first chunk of the realization
        if index == 0 or mode == 'wps-only':
            job.gw_job    = gw_job.submit( priority = priority, file_template = file_template )
        else:
            # if the chunk is not the first of the realization, 
            # gwsubmit has an argument, gw_job of the job before
            chunk_before_id = self.chunk_id - 1
            chunk_before    = self.realization.chunk.\
                              filter( Chunk.chunk_id == chunk_before_id ).one()
            job_before      = chunk_before.job.order_by( Job.id )[-1]
            id_job_before   = job_before.id          
            gw_job_before   = job_before.gw_job
            job.gw_job      = gw_job.submit( dep = gw_job_before, priority = priority,
                                             file_template = file_template )
        job.chunk_id = self.chunk_id
        job.run( rerun ) 
        self.job.append( job )

        # Update realization status
        if mode=='wrf':
            self.status = Chunk.Status.SUBMITTED
        else:
            self.status=Chunk.Status.WPS_SUBMITTED

    def cancel(self, hard = False ):
        """
        Delete jobs
        """
        logging.info('\t---> Canceling Chunk %d %s %s' % ( self.chunk_id,
                                                           datetime2datewrf(self.start_date),
                                                           datetime2datewrf(self.end_date) ) )
        l_jobs = self.job.filter( and_( Job.status != Job.Status.PREPARED, 
                                        Job.status != Job.Status.FINISHED,
                                        Job.status != Job.Status.FAILED,
                                        Job.status != Job.Status.CANCEL ) 
                                ).all()
        if not ( l_jobs ):
            logging.info( '\t\tThere are not jobs to cancel.' )
        else :
            for job in l_jobs :
                job.dryrun = self.dryrun
                job.cancel( hard )

    def set_priority(self, priority = 0 ):
        """
        Setting priority to jobs
        """
        logging.info('\t---> Setting priority Chunk %d %s %s' % ( self.chunk_id,
                                                                  datetime2datewrf(self.start_date),
                                                                  datetime2datewrf(self.end_date) ) )
        l_jobs = self.job.filter_by( status = Job.Status.SUBMITTED ).all()
        if not ( l_jobs ):
            logging.info( '\t\tThere are not jobs to set priority.' )
        else :
            for job in l_jobs :
                job.dryrun = self.dryrun
                job.set_priority( priority )
 
class JobCodeError():
    LOG_PATH             = 1
    COPY_APP             = 2
    APP_ERROR            = 3
    SOURCE_SCRIPT        = 4
    JOB_SHOULD_NOT_RUN   = 5
    LOCAL_PATH           = 6
    BINARY               = 7
    COPY_RST_FILE        = 8
    RESTART_MISMATCH     = 9
    COPY_FILE            = 10
    COPY_REAL_FILE       = 11
    COPY_BOUND           = 12
    NAMELIST_FAILED      = 13
    PREPROCESSOR_FAILED  = 14
    LINK_GRIB_FAILED     = 15
    UNGRIB_FAILED        = 16
    UNGRIB_PROCESSOR_FAILED  = 31
    METGRID_FAILED       = 17
    REAL_FAILED          = 18
    COPY_UPLOAD_WPS      = 19
    WRF_FAILED           = 20
    POSTPROCESSOR_FAILED = 21
    COPY_OUTPUT_FILE     = 22
   
class Job( object ):
    """
    A class to manage WRF4G jobs
    """
    dryrun = False

    Status = Enumerate( 'UNKNOWN', 'PREPARED', 'SUBMITTED','WPS_SUBMITTED', 'RELEASED', 'RUNNING', 'PENDING', 
                        'CANCEL', 'FAILED', 'FINISHED','WPS_FINISHED', 'CREATE_OUTPUT_PATH', 
                        'CONF_APP', 'DOWN_RESTART', 'DOWN_WPS', 'DOWN_BOUND', 'UNGRIB', 
                        'METGRID', 'REAL', 'UPLOAD_WPS', 'ICBCPROCESOR', 'WRF' )

    CodeError = JobCodeError()

    def set_status(self, status, timestamp = False ):
        """ 
        Save the status of the job and if it is a jobstatus of CHUNK STATUS and REALIZATION STATUS,
        change the status of the Chunk and the Realization and add an event. 
        """
        #Save job's status
        self.status = status
        if status == Job.Status.CANCEL :
            self.chunk.status = Chunk.Status.PREPARED
            self.chunk.realization.status = Realization.Status.PREPARED
        #if it is an status of the CHUNK STATUS 
        elif status in Chunk.Status and status != Chunk.Status.SUBMITTED :
            self.chunk.status = status
            #if it is an status of the REALIZATION STATUS 
            if status in Realization.Status and status != Realization.Status.SUBMITTED :
                # Last chunk has finished
                if ( status == Job.Status.FINISHED and \
                     self.chunk.chunk_id == self.chunk.realization.nchunks ) :
                    self.chunk.realization.status = Realization.Status.FINISHED
                # Finished chunk other than last
                elif ( status == Job.Status.FINISHED and \
                       self.chunk.realization.status != Realization.Status.FINISHED and \
                       self.chunk.realization.status != Realization.Status.FAILED ) :
                    self.chunk.realization.current_chunk = self.chunk.chunk_id + 1                    
                    self.chunk.realization.status = Realization.Status.SUBMITTED
                elif status != Job.Status.FINISHED :
                    self.chunk.realization.status = status
        #Add event
        events            = Events()
        events.job_status = status
        if not timestamp :
            events.timestamp  = datetime.utcnow()
        else :
            events.timestamp  = timestamp
        self.events.append( events ) 
        
    def run(self, rerun = False):
        """ 
        Send a gridway's job and save data in table Job
        id_chunk is an instance of Chunk
        """
        if not self.gw_restarted :
            self.gw_restarted = 0
        #Insert values default
        self.resource = '-'
        self.exitcode = '-'
        # Update status
        self.set_status( Job.Status.SUBMITTED ) 
   
    def cancel(self, hard = False ):
        """
        Delete a job
        """
        logging.info('\t\t---> Canceling Job %d' % self.gw_job ) 
        if not self.dryrun :
            GWJob().kill( self.gw_job, hard )
            self.set_status( Job.Status.CANCEL )

    def set_priority(self, priority = 0 ):
        """
        Set job priority
        """
        logging.info('\t\t---> Setting priority Job %d' % self.gw_job )
        if not self.dryrun :
            GWJob().set_priority( self.gw_job, priority )

class Events( object ) :
    """
    A class to manage job events
    """
    pass





#
# Copyright 2016 Universidad de Cantabria
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

"""
Get status and history and cancel jobs.

Usage: 
    wrf4g job list    [ --dbg ] [ --delay=<seconds> ] [ <job_id> ] 
    wrf4g job cancel  [ --dbg ] [ --hard ] <job_id>  
    wrf4g job log     [ --dbg ] <job_id>
    wrf4g job history [ --dbg ] <job_id> 
   
Arguments:
   <job_id>               Job identifier.

Options:
   --dbg                  Debug mode.
   --delay=<seconds>      Refresh experiment information every delay seconds.
   --hard                 Remove jobs from without synchronizing.
    
Commands:
   list                   Monitor jobs previously submitted.
   cancel                 Cancel jobs.
   log                    Keep track of a job.
   history                Get information about the execution history of a job.

Job field information:
   JID                    Job identification.
   DM                     Dispatch Manager state, one of: 
                                pend, hold, prol, prew, wrap, epil, canl, stop, migr, done, fail.
   EM                     Execution Manager state: pend, susp, actv, fail, done.
   START                  The time the job entered the system.
   END                    The time the job reached a final state (fail or done).
   EXEC                   Total execution time, includes suspension time in the remote queue system.
   XFER                   Total file transfer time, includes stage-in and stage-out phases.
   EXIT                   Job exit code.
   TEMPLATE               Filename of the job template used for this job.
   HOST                   Hostname where the job is being executed.
   HID                    Host identification.
   PROLOG                 Total prolog (file stage-in phase) time.
   WRAPPER                Total wrapper (execution phase) time.
   EPILOG                 Total epilog (file stage-out esphase) time.
   MIGR                   Total migration time.
   REASON                 The reason why the job left this host.
   QUEUE                  Queue name. 
"""

import logging
import sys
import time
from os.path                import join, exists
from drm4g.commands         import Daemon
from wrf4g.utils.command    import cls
from wrf4g.utils.gridwaylib import GWJob

def run( arg ) :
    logging.basicConfig( format = '%(message)s',
                         level  = logging.DEBUG if arg[ '--dbg' ] else logging.INFO,
                         stream = sys.stdout )
    try :
        daemon = Daemon( )
        if not daemon.is_alive() :
            raise Exception( 'DRM4G is stopped. ')
        gw_job = GWJob()
        if arg['list']:
            if not arg[ '--delay' ] :
                gw_job.list( None if not arg['<job_id>'] else arg['<job_id>'] )   
            else :
                try:
                    while True :
                        cls()
                        gw_job.list( None if not arg['<job_id>'] else arg['<job_id>'] )
                        time.sleep( int( arg[ '--delay' ] ) )
                except KeyboardInterrupt :
                    pass
        elif arg['history']:
            gw_job.history( arg['<job_id>'] )
        elif arg['log']:
            gw_job.log( arg['<job_id>'] )
        else :
            gw_job.kill( arg['<job_id>'], arg[ '--hard' ] )
    except KeyboardInterrupt :
        pass
    except Exception as err :
        logging.error( str( err ) )



#
# Copyright 2016 Universidad de Cantabria
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import os
import logging
import subprocess
from distutils import spawn


def cls():
    os.system("clear")


def which(command):
    """
    Locate commands
    """
    return spawn.find_executable(command)


def exec_cmd_advance(
    cmd,
    nohup=False,
    stdin=subprocess.PIPE,
    stdout=subprocess.PIPE,
    stderr=subprocess.STDOUT,
    env=os.environ,
):
    """
    Execute shell commands
    """
    logging.debug("Executing command ... " + cmd)
    cmd_to_exec = subprocess.Popen(
        cmd,
        shell=True,
        stdin=stdin,
        stdout=stdout,
        stderr=stderr,
        env=env,
        universal_newlines=True,
    )
    if not nohup:
        out, err = cmd_to_exec.communicate()
    else:
        out = err = ""
    return out, err


def exec_cmd(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=os.environ):
    logging.debug("Executing command ... " + cmd)
    p = subprocess.Popen(
        cmd,
        shell=True,
        stdout=stdout,
        stderr=stderr,
        env=env,
        universal_newlines=True
    )
    out, err = p.communicate()
    output = out.strip() + err.strip()
    return p.returncode, output


def os_stat(ifile):
    return os.stat(ifile)



#
# Copyright 2016 Universidad de Cantabria
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

def dict_compare(dict_one, dict_two ):
    """ 
    Compare two python dictionaries
    """
    dict_one_keys  = set( dict_one.keys() )
    dict_two_keys  = set( dict_two.keys() )
    intersect_keys = dict_one_keys.intersection( dict_two_keys  )
    added          = dict_one_keys - dict_two_keys
    removed        = dict_two_keys - dict_one_keys
    modified       = [ k for k in intersect_keys if dict_one[ k ] != dict_two[ k ] ]
    same           = set( k for k in intersect_keys if dict_one[ k ] == dict_two[ k ] )
    return ( added, removed, modified, same )

class dict2obj(dict):
    """
    Class to convert a dictionary to an object
    """
    def __init__(self, dictionary, default=None):
        self.__dictionary = dictionary
        self.__default = default
        super(self.__class__, self).__init__(dictionary)

    def __getattr__(self, key ):
        if key.startswith('__'):
            raise AttributeError
        if key in self.__dictionary :
            val = self.__dictionary[ key ]
            if isinstance(val, dict):
                val = self.__class__( val )
            setattr(self, key, val )
            return val
        return self.__default

    def __missing__(self, key):
        return False

class Enumerate( frozenset ):

    def __new__(cls, *args):
        return frozenset.__new__(cls, args)

    def __getattr__(self, name):
        if name in self:
            return name
        else:
            raise AttributeError("No '%s' in enumeration" % name )




#
# Copyright 2016 Universidad de Cantabria
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

"""
Manage computing resources on WRF4G.
    
Usage: 
    wrf4g resource [ --dbg ] [ list | edit | check ]
    
Options:
    --dbg     Debug mode.
    
Commands:
    list      Show resources available.    
    edit      Configure resouces.
    check     Check out if configured resources are accessible.
"""

import logging
import sys
from drm4g.core.configure import Configuration
from drm4g.commands       import Daemon, Resource

def run( arg ) :
    # El logger no funcionaba porque el import logging se hace varias veces y coge la primera. Hacemos un reload para que funcione
    # https://stackoverflow.com/questions/20240464/python-logging-file-is-not-working-when-using-logging-basicconfig
    from imp import reload
    reload(logging)
    logging.basicConfig( format = '%(message)s',
                     level  = logging.DEBUG if arg[ '--dbg' ] else logging.INFO,
                     stream = sys.stdout )
                     
    try :
        config = Configuration()
        daemon = Daemon()
        if not daemon.is_alive() :
           raise Exception( 'DRM4G is stopped.' )
        resource = Resource( config )
        if arg[ 'edit' ] :
            resource.edit()
        elif arg[ 'check' ] :
            resource.check_frontends( )
        else :
            resource.list()
    except KeyboardInterrupt :
        pass
    except Exception as err :
        logging.error( str( err ) )




#
# Copyright 2016 Universidad de Cantabria
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

"""
Configure DRM4G daemon, scheduler, database and logging parameters.
                
Usage:  
   wrf4g conf ( daemon | sched | logger | database ) [ --dbg ]

Options:
   --dbg    Debug mode
"""

import logging
import sys
from wrf4g            import DB4G_CONF
from drm4g            import DRM4G_GWD_CONF, DRM4G_SCHED_CONF, DRM4G_LOGGER_CONF
from wrf4g.utils.file import edit_file

def run( arg ) :
    logging.basicConfig( format = '%(message)s', 
                         level  = logging.DEBUG if arg[ '--dbg' ] else logging.INFO,
                         stream = sys.stdout )
    if arg[ 'daemon' ] :
        conf_file = DRM4G_GWD_CONF
    elif arg[ 'logger' ]:
        conf_file = DRM4G_LOGGER_CONF
    elif arg[ 'database' ]:
        conf_file = DB4G_CONF
    else :
        conf_file = DRM4G_SCHED_CONF
    edit_file( conf_file )



#
# Copyright 2016 Universidad de Cantabria
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import os
import logging
import fortran_namelist as fn
from os.path          import exists
from wrf4g.utils.time import datetime2datewrf

def get_ptop():
    shcmd = "ncdump -v PRES $(\ls met_em.d??.????-??-??_??[:_]00[:_]00.nc | head -1) | tail -2 | head -1 | tr -d ',;' | awk '{printf $1}'"
    out=os.popen(shcmd).read().strip()
    return int(out)

def fix_ptop( namelist_input ):
    top_press = get_ptop()
    nmli = fn.WrfNamelist( namelist_input )
    try :
        specp_top = int( nmli.getValue( "p_top_requested" )[0] )
    except Exception :
        if top_press > 5000 :
            logging.info( "Default p_top is 5000., but your input files only reach %d. Fixing..." % top_press )
            nmli.setValue( "p_top_requested ", top_press, "domains" )
        else :
            logging.info( "Default p_top will be 5000." )
            nmli.setValue( "p_top_requested ", 5000, "domains" )
    else :
        if specp_top < top_press :
            logging.info( "Specified p_top is %d, but your input files only reach %d. Fixing..." % ( specp_top, top_press ) )
            nmli.setValue( "p_top_requested", top_press )
    nmli.wrfCheck()
    nmli.overWriteNamelist()

def get_num_metgrid_levels():
    shcmd = "ncdump -h $(\ls -1 met_em*.nc | head -1) | grep 'num_metgrid_levels =' | sed -e 's/^\t//' | tr '=;' '  ' | awk '{print $2}'"
    out=os.popen(shcmd).read().strip()
    return int(out)

def get_num_metgrid_soil_levels():
    shcmd = "ncdump -h $(\ls -1 met_em*.nc | head -1) | grep 'NUM_METGRID_SOIL_LEVELS' | sed -e 's/^\t//' | tr '=:;' '   ' | awk '{print $2}'"
    out=os.popen(shcmd).read().strip()
    return int(out)

def get_time_step(coarse_dx, factor):
    HOUR_DIVISORS = [
        1,2,3,4,5,6,8,9,10,12,15,16,18,20,24,25,30,36,40,45,48,50,60,
        72,75,80,90,100,120,144,150,180,200,225,240,300,360,400,450,600,
        720,900,1200
        ]
    tstep = int(factor*coarse_dx/1000)
    # Added list to fix execution in python 3
    ival = list(map(lambda x: x<=tstep, HOUR_DIVISORS)).index(0) - 1
    return HOUR_DIVISORS[ival]

def get_latlon_dx(start_date, dom):
    #  Try to get dx from the met_em or wrfinput files. Only
    #  required for lat-lon grids, otherwise it is available
    #  in the namelist.wps file
    dxfile=None
    file_name = "met_em.%s.%s.nc" % ( dom, datetime2datewrf( start_date ) )
    if exists( file_name ) :
        dxfile = file_name
        logging.info("Reading dx from %s" % file_name)
    else:
        file_name = "wrfinput_%s" % dom
        dxfile = file_name
        logging.info("Reading dx from %s" % file_name)
    if dxfile:
        shcmd = "ncdump -h %s | grep 'DX =' | sed -e 's/^\t//' | tr '=;' ' ' | awk '{printf \"%%f\", $2}'" % dxfile
        rval = round(float(os.popen(shcmd).read().strip()), 4)
    else:
        raise Exception('get_latlon_dx: no met_em or wrfinput file found')
    return rval

def wps2wrf( namelist_wps, namelist_input, sdate, edate, maxdom, chunk_is_restart, timestep_dxfactor='6') :
    nmlw = fn.FortranNamelist( namelist_wps )
    nmli = fn.WrfNamelist( namelist_input )
    nmli.setValue("max_dom", maxdom)
    for var in ["run_days", "run_hours", "run_minutes", "run_seconds"]:
        nmli.setValue(var, 0)
    nmli.setMaxDomValue("start_year",  sdate.year)
    nmli.setMaxDomValue("start_month", sdate.month)
    nmli.setMaxDomValue("start_day",   sdate.day)
    nmli.setMaxDomValue("start_hour",  sdate.hour)
    nmli.setMaxDomValue("end_year",    edate.year)
    nmli.setMaxDomValue("end_month",   edate.month)
    nmli.setMaxDomValue("end_day",     edate.day)
    nmli.setMaxDomValue("end_hour",    edate.hour)
    vars_to_copy = ["parent_grid_ratio", "i_parent_start", "j_parent_start",
                    "e_we", "e_sn", "interval_seconds"]
    for var in vars_to_copy:
        nmli.setValue(var, nmlw.getValue(var))
    nmli.setValue("parent_time_step_ratio", nmlw.getValue("parent_grid_ratio"))
    if exists("met_em.d01.%s.nc" % datetime2datewrf( sdate ) ):
        # If there are met_em files, we need to run real.exe. Otherwise, we
        # cannot get enough info (num_metgrid_*levels) to run real.exe
        nmli.setValue("num_metgrid_levels", get_num_metgrid_levels())
        nmli.setValue("num_metgrid_soil_levels", get_num_metgrid_soil_levels())
    #
    #  Compute the grid spacings. Read them from met_em files if the projection is lat-lon.
    #
    nmli.setValue("grid_id", list(range(1, maxdom+1)))
    # Update parant_id in the namelist
    nmli.setValue("parent_id", nmlw.getValue("parent_id"))

    alldx = []
    for idom in range(1,maxdom + 1):
        thisdx = get_latlon_dx(sdate, "d0%i" % idom)
        alldx.append(thisdx)

    nmli.setValue("dx", alldx)
    nmli.setValue("dy", alldx) # May be an issue for global WRF
    #
    # Compute the time step. 
    #
    if timestep_dxfactor.startswith("manual:"):
        nmli.setValue("time_step", int(timestep_dxfactor[7:]))
    elif timestep_dxfactor.startswith("adaptive:"):
        nmli.setValue("use_adaptive_time_step", ".true.", "domains")
    else:
        nmli.setValue("time_step",
            get_time_step(nmli.getValue("dx")[0], eval(timestep_dxfactor)))
    nmli.setValue("restart", chunk_is_restart)
    #
    #  Trim, check, overwrite the file and ... we are done!
    #
    #nmli.trimMaxDom()
    nmli.wrfCheck()
    #nmli.extendMaxDomVariables()
    nmli.overWriteNamelist()



#!/usr/bin/env python
#
# fortran_namelist
#
# ------------------------------------------------------------------------
# Copyleft 2009, Jesus Fernandez <jesusff IN THE DOMAIN gmail DOT com>
#
# Santader Meteorology Group, Universidad de Cantabria, Spain
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------

import sys
import logging

def listify(item):
  if type(item) == type([]):
    return [ elem for elem in item ]
  else:
    return [item,] 

def flat_list(l):
  rval = []
  for item in l:
    if type(item) == type([]):
      rval.extend(item)
    else:
      rval.append(item)
  return rval

def math_expansion(expr):
  try:
    exec("rval=%s" % expr)
    return "%s" % rval
  except:
    return expr

def asterisk_expansion(expr):
  if '*' in expr:
    [mult, value] = expr.split('*')
    rval = int(mult)*[value,]
    return list(map(math_expansion, rval))
  else:
    return expr

def coerce_value_list(vlist, math_exp=True):
  if math_exp:
    vlist = list(map(math_expansion, vlist))
  else:
    vlist = list(map(asterisk_expansion, vlist))
    vlist = flat_list(vlist)
  try:
    values=[int(element) for element in vlist]
  except ValueError:
    try:
      values=[float(element) for element in vlist]
    except ValueError:
      values=[value.strip() for value in vlist]
  return values

class FortranNamelistRecord:
  def __init__(self, name):
    self.name = name
    self.data = {}
    self.ordered_keys = []
  def setValue(self, variable, value):
    self.data[variable] = listify(value)
    if not variable in self.ordered_keys:
      self.ordered_keys.append(variable)
  def delVariable(self, variable):
    if variable in self.ordered_keys:
      del self.data[variable]
      self.ordered_keys.remove(variable)
  def appendValue(self, variable, value):
    if not variable in self.ordered_keys:
      self.ordered_keys.append(variable)
      self.data[variable]=[]
    self.data[variable].extend(listify(value))
  def hasVariable(self, var):
    return var in self.ordered_keys
  def printRecord(self, sorted=False):
    rval = "&%s\n" % self.name
    keys = self.ordered_keys
    if sorted:
      keys.sort()
    for key in keys:
      thisvar = "  %-24s =" % key
      for item in self.data[key]:
        if type(item) == type("hello"):
          if not item.startswith("'") and not item[:2].upper() in [".T", ".F"]:
            item = "'%s'" % item
        if len(thisvar) < 75:
          thisvar += " %s," % item
        else:
          rval += "%s\n" % thisvar
          thisvar = 28*" "+" %s," % item
      rval += "%s\n" % thisvar
    rval += "/\n"
    return rval
  def __str__(self):
    self.printRecord(sorted=False)
  def __getitem__(self, item):
    return self.data[item]

class FortranNamelist:
  def __init__(self, namelist_file, math_exp=True):
    """read contents of namelist file and return dictionary containing all options
    
    Created 20/01/08 by Thom Chubb.
    Modified 20/01/08 by Thom Chubb and Valerio Bisignesi
    Modified 20/05/09 by Jesus Fernandez 
      Implemented as a class
      Added support for multi-line variables
    """
    self.namelist_file = namelist_file
    self.math_exp = math_exp
    fid=open(namelist_file, 'r')
    data = fid.readlines()
    fid.close()
    self.record_dict={}
    self.ordered_records=[]
    is_comment = True
    for line in data:
      if line.strip().startswith('!'): continue
      if '&' in line:
        # Then this line is a namelist title
        is_comment=False
        current_label = line.strip().lstrip('&')
        current_label = current_label.lower()
        self.record_dict[current_label] = FortranNamelistRecord(current_label)
        if not current_label in self.ordered_records:
          self.ordered_records.append(current_label)
      elif line.strip() == "/":
        # Then lines following this are comments until the next '&'
        is_comment=True
      elif '=' in line:
        # Then this line contains variable information to be stored
        if not is_comment:
          variable,values = line.split('=')
          variable = variable.lower()
          values = values.strip().rstrip(',')
          if values.startswith("'") and values.endswith("'") and not "'" in values[1:-1]:
            # This is a single string with comma-separated values. Do not interpret it as 
            # as comma-separated strings.
            values = values[1:-1]
          else:
            values = coerce_value_list(values.split(','), self.math_exp)
          self.record_dict[current_label].setValue(variable.strip(), values)
      else:
        if not is_comment:
          # This line contains variable information to be added to the last variable read
          values = line.strip().rstrip(',')
          values = coerce_value_list(values.split(','), self.math_exp)
          self.record_dict[current_label].appendValue(variable.strip(), values)
  def __getitem__(self, item):
    return self.record_dict[item]
  def printNamelist(self, record=None, sorted=False):
    rval = ""
    if record:
      rval += "%s\n" % self.record_dict[record].printRecord(sorted)
    else:
      records = self.ordered_records
      if sorted:
        records.sort() 
      for rec in records:
        rval += "%s\n" % self.record_dict[rec].printRecord(sorted)
    return rval
  def setValue(self, variable, value, record=""):
    if record:
      if record not in self.record_dict:
        self.record_dict[record] = FortranNamelistRecord(record)
        if not record in self.ordered_records:
          self.ordered_records.append(record)
      self[record].setValue(variable, value)
    else:
      gotit=False
      for rec in self.ordered_records:
        if self[rec].hasVariable(variable):
          self[rec].setValue(variable, value)
          gotit=True
          break
      if not gotit:
        raise KeyError( "The variable '%s' was not found and no record was specified!" % variable )
  def getValue(self, variable, record=""):
    if record:
      return self[record][variable]
    else:
      gotit=False
      for rec in self.ordered_records:
        if self[rec].hasVariable(variable):
          return self[rec][variable]
          gotit=True
          break
      if not gotit:
        raise KeyError( "The variable '%s' was not found." % variable )
  def hasVariable(self, variable, record=""):
    if record:
      return self[record].hasVariable(variable)
    else:
      rval = False
      for rec in self.ordered_records:
        if self[rec].hasVariable(variable):
          rval = True
          break
      return rval
  def variableList(self):
    rval = []
    for rec in self.ordered_records:
      rval.extend(self[rec].ordered_keys)
    return rval
  def appendValue(self, variable, value, record=""):
    if record:
      self[record].appendValue(variable, value)
    else:
      gotit=False
      for rec in self.ordered_records:
        if self[rec].hasVariable(variable):
          self[rec].appendValue(variable, value)
          gotit = True
          break
      if not gotit:
        raise KeyError( "The variable '%s' was not found and no record was specified!" % variable )
  def delVariable(self, variable, record=""):
    if record:
      self[record].delVariable(variable)
    else:
      for rec in self.ordered_records:
        if self[rec].hasVariable(variable):
          self[rec].delVariable(variable)
          break
  def overWriteNamelist(self):
    fid = open(self.namelist_file, 'w')
    fid.write(self.printNamelist())
    fid.close()

class WrfNamelist(FortranNamelist):
  # Get an updated list for your WRF version with:
  # grep '^rconfig' Registry |awk '$5=="max_domains"{printf "  %s,\n", tolower($8)}' | sort | uniq
  # "feedback" wrongly appears as max_dom var (how can this be posible?)
  # and was removed by hand. "frames_per_outfile" and "history_interval"
  # were missing and added by hand.
  # Last updated from WRF v3.7 (Carlos Blanco, 20150717)
  MAX_DOM_VARIABLES = [
    "allowed", "auxhist1_interval", "auxhist2_interval", "auxhist3_interval",
    "auxhist4_interval", "auxhist5_interval", "auxhist6_interval", 
    "auxhist7_interval", "auxhist8_interval", "auxhist9_interval", 
    "auxhist10_interval", "auxhist11_interval", "bdyfrq", "bl_pbl_physics", 
    "bldt", "c_k", "c_s", "cen_lat", "cen_lon", "chem_adv_opt", "chem_opt", 
    "coriolis2d", "cu_diag", "cu_physics", "cudt", "cycle_x", "cycle_y", 
    "dampcoef", "decoupled", "dfi_stage", "diff_6th_factor", "diff_6th_opt", 
    "do_coriolis", "do_curvature", "do_gradp", "dt", "dx", "dy", "e_sn", 
    "e_vert", "e_we", "emdiv", "end_day", "end_hour", "end_minute", "end_month", 
    "end_second", "end_year", "epssm", "fdda_end", "fdda_start", "fgdt", 
    "fgdtzero", "fine_input_stream", "frames_per_outfile", "frames_per_auxhist1", 
    "frames_per_auxhist2","frames_per_auxhist3", "frames_per_auxhist4", 
    "frames_per_auxhist5", "frames_per_auxhist6","frames_per_auxhist7",
    "frames_per_auxhist8", "frames_per_auxhist9","frames_per_auxhist10",
    "frames_per_auxhist11", "frames_per_auxhist12","frames_per_auxhist13",
    "frames_per_auxhist14", "frames_per_auxhist15","frames_per_auxhist16",
    "frames_per_auxhist17", "frames_per_auxhist18","frames_per_auxhist19",
    "frames_per_auxhist20","frames_per_auxhist21", "frames_per_auxhist22", 
    "frames_per_auxhist23","frames_per_auxhist24", "frames_per_auxhist25", 
    "gmt", "grav_settling", "grid_fdda", "grid_id", 
    "gsmdt", "history_interval","history_interval_mo", "history_interval_d", 
    "history_interval_h", "history_interval_m", "history_interval_s",
    "h_mom_adv_order", "h_sca_adv_order", "i_parent_start", "id", "input_from_file",
    "input_from_hires", "iofields_filename", "isice", "islake", "isoilwater",
    "isurban", "iswater", "j_parent_start", "julday", "julyr", "khdif",
    "kvdif", "m_opt", "map_proj", "max_step_increase_pct", "max_time_step",
    "min_time_step", "mix_full_field", "mix_isotropic", "mix_upper_bound",
    "moad_cen_lat", "moad_grid_ratio", "moad_time_step_ratio", "specified",
    "moist_adv_dfi_opt", "moist_adv_opt", "mp_physics", "mp_physics_dfi",
    "naer", "nested", "non_hydrostatic", "obs_coef_mois", "obs_coef_pstr",
    "obs_coef_temp", "obs_coef_wind", "obs_ionf", "obs_no_pbl_nudge_q",
    "obs_no_pbl_nudge_t", "obs_no_pbl_nudge_uv", "obs_nudge_mois",
    "obs_nudge_opt", "obs_nudge_pstr", "obs_nudge_temp", "obs_nudge_wind",
    "obs_prt_freq", "obs_rinxy", "obs_twindo", "open_xe", "open_xs", "open_ye",
    "open_ys", "parent_grid_ratio", "parent_id", "parent_time_step_ratio",
    "periodic_x", "periodic_y", "pert_coriolis", "polar", "pole_lat",
    "pole_lon", "prec_acc_dt", "progn", "pxlsm_smois_init", "pxlsm_soil_nudge",
    "ra_lw_physics", "ra_sw_physics", "radt", "s_sn", "s_vert", "s_we",
    "scalar_adv_opt", "sf_sfclay_physics", "sf_surface_physics",
    "sf_urban_physics", "shcu_physics", "slope_rad", "smdiv", "specified",
    "stand_lon", "start_day", "start_hour", "start_minute", "start_month",
    "start_second", "start_year", "starting_time_step", "stencil_half_width",
    "swap_x", "swap_y", "symmetric_xe", "symmetric_xs", "symmetric_ye",
    "symmetric_ys", "target_cfl", "target_hcfl", "time_step_sound", "tke_adv_opt", 
    "tke_drag_coefficient", "tke_heat_flux", "tke_upper_bound", "top_lid", 
    "top_radiation", "topo_shading", "topo_wind","tracer_adv_opt", "tracer_opt",
    "true_lat1", "true_lat2", "v_mom_adv_order", "v_sca_adv_order", "zdamp", "ztop"
  ]
  NAMELIST_RECORDS = [ 
    "bdy_control", "chem", "dfi_control", "diags", "domains", "dynamics", "scm",
    "fdda", "fire", "grib2", "logging", "namelist_01", "namelist_02", "time_control", 
    "namelist_03", "namelist_04", "namelist_05", "namelist_quilt", "physics",
    "noah_mp", "tc"
  ]
  def __init__(self, namelist_file, math_exp=True):
    FortranNamelist.__init__(self, namelist_file, math_exp)
  def setMaxDomValue(self, variable, value, record=""):
    mxd = self['domains']['max_dom'][0]
    if record:
      self[record].setValue(variable, coerce_value_list(mxd*[value,], self.math_exp))
    else:
      gotit=False
      for rec in self.ordered_records:
        if self[rec].hasVariable(variable):
          self[rec].setValue(variable, coerce_value_list(mxd*[value,], self.math_exp))
          gotit=True
          break
      if not gotit:
        raise KeyError( "The variable was not found and no record was specified!" )
  def checkMaxDomPatterns(self, var):
    prefixes = ["start_", "end_"]
    suffixes = ["_inname", "_outname", "_interval"]
    rval = False
    for prefix in prefixes:
      if var.startswith(prefix):
        rval = True
        break
    for suffix in suffixes:
      if var.endswith(suffix):
        rval = True
        break
    return rval
  def trimMaxDom(self, ncols=None):
    mxd = self['domains']['max_dom'][0]
    if ncols:
       mxd = ncols
    for var in self.variableList():
      if ncols or var in self.MAX_DOM_VARIABLES:# or self.checkMaxDomPatterns(var):
        self.printWrfWarning('Trimming variable %s.' % var)
        self.setValue(var, self.getValue(var)[:mxd])
  def printWrfError(self, message):
    logging.error("WRF Check Error: %s" % message)
  def printWrfWarning(self, message):
    logging.warn("WRF Check Warning: %s" % message)
  def wrfCheck(self):
    """
    Check for some recomendations/mandatory WRF specific issues in the namelist.
    """
    error = 0
    for record in list(self.record_dict.keys()):
      if not record in self.NAMELIST_RECORDS :
        self.printWrfError( "'%s' section does not exist" % record )
        error = error + 1
    if self.hasVariable('eta_levels', 'domains'):
      if not self.hasVariable('e_vert', 'domains'): 
        self.printWrfError( "Selected eta_levels but e_vert was not set" )
        error = error + 1
      elif len( self.getValue('eta_levels', 'domains') ) != self.getValue('e_vert', 'domains')[0]:
        self.printWrfError( "eta_levels are not %s, there are %s" % ( self.getValue('e_vert', 'domains')[0],
                                                                     len( self.getValue('eta_levels', 'domains') ) ) )
        error = error + 1
    tsratio = self.getValue('time_step')[0] * 1000 / self.getValue('dx')[0]
    if   tsratio > 6: self.printWrfWarning("Time step is larger than 6 times dx (%f)" % tsratio)
    elif tsratio < 5: self.printWrfWarning("Time step is shorter than 5 times dx (%f)" % tsratio)
    raratio = self.getValue('radt')[0] * 1000 / self.getValue('dx')[0]
    if   raratio > 1.1: self.printWrfWarning("radt is larger than dx (%f)" % raratio)
    if   raratio < 0.9: self.printWrfWarning("radt is shorter than dx (%f)" % raratio)
    #
    # SST update
    #
    if self.hasVariable('sst_update', 'physics'):
      if self.getValue('sst_update','physics')[0] == 1 and not self.hasVariable('auxinput4_inname', 'time_control'):
         self.printWrfWarning("sst_update enabled but auxinput4_inname not defined! Fixing... (check interval!)")
         self['time_control'].setValue('auxinput4_inname', "wrflowinp_d<domain>")
         self['time_control'].setValue('io_form_auxinput4', 2)
         self['time_control'].setValue('auxinput4_end_h', 0)
         self.setMaxDomValue('auxinput4_interval', 360, 'time_control')
    #
    # CAM radiation
    #
    if self.getValue('ra_lw_physics')[0] == 3 and not self.hasVariable('paerlev', 'physics'):
      self.printWrfWarning('CAM radiation selected but paerlev/levsiz/cam_abs_dim1/cam_abs_dim2 was not set. Fixing...')
      self['physics'].setValue('paerlev', 29)
      self['physics'].setValue('levsiz', 59)
      self['physics'].setValue('cam_abs_dim1', 4)
      self['physics'].setValue('cam_abs_dim2', self.getValue('e_vert')[0])
    if self.getValue('ra_lw_physics')[0] == 3 and self.getValue('cam_abs_dim2')[0]!=self.getValue('e_vert')[0]:
      self.printWrfWarning('cam_abs_dim2 not set to e_vert. Fixing...')
      self['physics'].setValue('cam_abs_dim2', self.getValue('e_vert')[0])
    #
    # PBL issues
    #
    if self.getValue('bl_pbl_physics')[0] == 1 and self.getValue('sf_sfclay_physics')[0] != 1:
      self.printWrfWarning('YSU PBL selected but the surface layer selected is not 1. Fixing...')
      self['physics'].setValue('sf_sfclay_physics', 1)
    #
    # LSM issues
    #
    if self.getValue('sf_surface_physics')[0] == 7 and self.getValue('num_soil_layers')[0] != 2:
      self.printWrfWarning('Pleim Xiu LSM selected but the soil levels are not 2. Fixing...')
      self['physics'].setValue('num_soil_layers', 2)
    if self.getValue('sf_surface_physics')[0] == 3 and self.getValue('num_soil_layers')[0] != 6:
      self.printWrfWarning('RUC LSM selected but the soil levels are not 6. Fixing...')
      self['physics'].setValue('num_soil_layers', 6)
    if self.getValue('sf_surface_physics')[0] == 2 and self.getValue('num_soil_layers')[0] != 4:
      self.printWrfWarning('Noah LSM selected but the soil levels are not 4. Fixing...')
      self['physics'].setValue('num_soil_layers', 4)
    if self.getValue('sf_surface_physics')[0] == 1 and self.getValue('num_soil_layers')[0] != 5:
      self.printWrfWarning('Simple soil selected but the soil levels are not 5. Fixing...')
      self['physics'].setValue('num_soil_layers', 5)
    return error
  def extendMaxDomVariables(self):
    #
    # Provide enough columns in max_dom variables. Risky stuff..
    #
    for var in self.MAX_DOM_VARIABLES:
      if self.hasVariable(var):
        theval = self.getValue(var)
        if len(theval) < self.getValue("max_dom")[0]:
          self.printWrfWarning('Variable %s = %s requires as many entries as domains.' % (var, theval))
          if var == "grid_id":
            self.setValue(var, list(range(1, self.getValue("max_dom")[0]+1))) 
          else:
            self.setValue(var, theval + (self.getValue("max_dom")[0]-len(theval))*[theval[-1],])
          self.printWrfWarning('Filling with last domain entry!! -> %s = %s' % (var, self.getValue(var)))




#
# Copyright 2016 Universidad de Cantabria
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

"""
Check DRM4G daemon and ssh-agent. 
    
Usage: 
    wrf4g status [ --dbg ] 
   
Options:
   --dbg    Debug mode.
"""

import logging
import sys
from drm4g.commands       import Daemon, Agent

def run( arg ) :
    try:
        # El logger no funcionaba porque el import logging se hace varias veces y coge la primera. Hacemos un reload para que funcione
        # https://stackoverflow.com/questions/20240464/python-logging-file-is-not-working-when-using-logging-basicconfig
        from imp import reload
        reload(logging)
        logging.basicConfig( format = '%(message)s',
                         level  = logging.DEBUG if arg[ '--dbg' ] else logging.INFO,
                         stream = sys.stdout )
        Daemon().status()
        Agent().status()
    except KeyboardInterrupt :
        pass
    except Exception as err :
        logging.error( str( err ) )




#
# Copyright 2016 Universidad de Cantabria
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import os
import sys
import logging
from os.path             import dirname, abspath, join, exists
from wrf4g               import WRF4G_DEPLOYMENT_DIR, WRF4G_DIR
from wrf4g.utils.command import exec_cmd
from drm4g               import DRM4G_DIR

class GWJob( object ):
    """
    Class to use DRM4G 
    """
    def create_template(self,name,directory,arguments,np=1,req='',environ='',inputsandbox='',outputsandbox=''):
        """
        Create template
        """
        ftemplate = join( directory,  name + ".gw" )
        f = open(ftemplate,'w')
        template="""NAME = %s
EXECUTABLE   = /usr/bin/env python3
ARGUMENTS    = ./bin/wrf_wrapper.py %s
WRAPPER      = %s/etc/wrf4g_wrapper.sh
INPUT_FILES  = %s
OUTPUT_FILES = %s
REQUIREMENTS = %s
ENVIRONMENT  = %s
RESCHEDULE_ON_FAILURE = no
NP           = %d""" % (name,arguments,WRF4G_DIR,inputsandbox,outputsandbox,req,environ,np)
        f.write(template)
        f.close()
        return ftemplate
    
    def submit(self, dep = None, priority = 0, type_dep = "afterok", file_template = "job.gw" ):
        depend = "-d %s -r %s" % ( dep, type_dep ) if dep != None else '-o '
        cmd = "GW_LOCATION=%s gwsubmit -p %d -v %s -t %s" % ( DRM4G_DIR, priority, 
                                                  depend, file_template )
        code, out = exec_cmd( cmd )
        logging.debug( out )
        if code :
            raise Exception( out )
        return out[ 8: ]
   
    def history(self, job_id ):
        cmd = 'GW_LOCATION=%s gwhistory %s' % ( DRM4G_DIR, job_id )
        code, out = exec_cmd( cmd )
        logging.info( out )

    def log(self, job_id ):
        directory = join( WRF4G_DIR ,
                          'var' ,
                          '%d00-%d99' % ( int(int(float( job_id ))/100) , int(int(float( job_id ))/100) ) ,
                          job_id ,
                          'job.log' )
        if not exists( directory ) :
            raise Exception( 'There is not a log available for this job.')
        cmd = 'cat %s' % ( directory )
        code, out = exec_cmd( cmd )
        logging.info( out )
 
    def list(self, job_id = None ):
        cmd = 'GW_LOCATION=%s gwps -o Jsetxjh ' % DRM4G_DIR
        if job_id :
            cmd = cmd + job_id 
        code, out = exec_cmd( cmd )
        logging.info( out )

    def release(self, job_id ):
        cmd = "GW_LOCATION=%s gwkill -l %s" % ( DRM4G_DIR, str( job_id ) )
        code, out = exec_cmd( cmd )
        if code :
            logging.info( out )

    def kill(self, job_id, hard = False ):
        cmd = "GW_LOCATION=%s gwkill %s %s" % ( DRM4G_DIR, '-9' if hard else '', str( job_id ) )
        code, out = exec_cmd( cmd )
        if code :
            logging.info( out )
                
    def set_priority(self, job_id, priority ):
        cmd = "GW_LOCATION=%s gwkill -p %d %s" % ( DRM4G_DIR, priority, str(job_id ) )
        code, out = exec_cmd( cmd )
        logging.info( out )




#
# Copyright 2016 Universidad de Cantabria
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

"""
Manage WRF4G realizations. 
    
Usage: 
     wrf4g rea <name> submit       [ --dbg ] [ --dry-run ] [ --priority=<value> ] [ --rerun ] [ --wps-only ] [ --mode=<value> ] [ <first_ch> [ <last_ch> ] ]
     wrf4g rea <name> status       [ --dbg ] [ --delay=<seconds> ] [ --verbose ] [ --show-chunks ]
     wrf4g rea <name> info
     wrf4g rea <name> log          [ --dbg ] [ --dir=<directory> ] <chunk_id>
     wrf4g rea <name> set-priority [ --dbg ] [ --dry-run ] <priority>
     wrf4g rea <name> cancel       [ --dbg ] [ --dry-run ] [ --hard ]
     wrf4g rea <name> set-restart  [ --dbg ] <date>
     wrf4g rea <name> get-restart  
   
Options:
    --dbg                  Debug mode.
    -n --dry-run           Dry run.
    --rerun                Force to run although the realization has finished.
    --wps-only             Only run wps, not wrf 
    -P --priority=<value>  Fix-priority for scheduling [default: 0].
    --delay=<seconds>      Refresh experiment information every delay seconds.    
    -d --dir=<directory>   Directory to unpack log files [default: ./].
    --hard                 Remove jobs from without synchronizing.
    -m --mode=<value>      0 (default): run whole workflow, 1: only WPS and real, 2: only WRF [default: 0]
    --show-chunks               Show advanced information about the chunks status
  
Commands:
    submit                 Submit the realization.       
    status                 Check the status of a realization showing computing resources, 
                           job identifier and exit codes (SEE EXIT CODES).
    info                   Get information about configuration.
    log                    Get log files from a chunk.
    set-priority           Change the scheduling priority of any job releted to the realization. 
                           The priority must be in range [0,20], and the default value is 0. 
                           When a job gets a priority of 20, it becomes an urgent job, and it is 
                           dispatched as soon as possible passing all the scheduling policies. 
    cancel                 Cancel the realization by killing its jobs.
    set-restart            Modify realization restart date.
    get-restart            Show realization restart date.

EXIT CODES
    1  : Error creating log directory        
    2  : Error copying apps            
    3  : Error app type does not exist            
    4  : Error executing source script       
    5  : Job already executed 
    6  : Error creating directory to simulate
    7  : Error finding WRF binaries 
    8  : Error copying restart files        
    9  : There is a mismatch in the restart date   
    10 : Error copying files    
    11 : Error downloading WPS files    
    12 : Error copying boundaries           
    13 : Error modifying namelist
    14 : Error executing PREPROCESSOR
    15 : Error linking GRIB files     
    16 : Error executing UNGRIB
    17 : Error executing METGRID       
    18 : Error executing REAL
    19 : Error uploading WPS files      
    20 : Error executing WRF
    21 : Error executing POSTPROCESSOR 
    22 : Error copying output file    
    23 : Job killed by the system 
    255: Unexpected error
"""

import logging
import sys
import time
from wsgiref.handlers import read_environ
from sqlalchemy.orm.exc   import NoResultFound
from sqlalchemy.exc       import OperationalError
from wrf4g.utils.command  import cls
from wrf4g.db             import get_session
from wrf4g.core           import Realization

def run( arg ) :
    logging.basicConfig( format = '%(message)s',
                         level  = logging.DEBUG if arg[ '--dbg' ] else logging.INFO,
                         stream = sys.stdout )
    # create a session to connect with the database
    session = get_session()
    try :
        try :
            rea = session.query( Realization ).\
                    filter( Realization.name   == arg[ '<name>' ], 
                            Realization.exp_id != None ).one()
            rea.dryrun = arg[ '--dry-run' ]
        except NoResultFound :
            raise Exception( "'%s' realization does not exist" % arg[ '<name>' ] )
        else :
            if arg[ 'submit' ] :
                if arg['--wps-only']:
                    rea.run_wps(first_chunk_run = arg[ '<first_ch>' ], 
                            last_chunk_run  = arg[ '<last_ch>' ], 
                            rerun           = arg[ '--rerun' ],
                            priority        = int( arg[ '--priority' ] ), 
                            )
                else:
                    rea.run( first_chunk_run = arg[ '<first_ch>' ], 
                            last_chunk_run  = arg[ '<last_ch>' ], 
                            rerun           = arg[ '--rerun' ],
                            priority        = int( arg[ '--priority' ] ) 
                            )
            elif arg[ 'status' ] :
                if arg['--show-chunks']: 
                    showchunks=1
                else:
                    showchunks=0
                if not arg[ '--delay' ] :
                    if showchunks == 0:
                        rea.status_header( )
                    else:
                        rea.status_chunk_header()

                    rea.get_status(showchunks)
                else :
                    try:
                        while True :
                            cls()
                            rea.status_header()
                            rea.get_status(showchunks)
                            time.sleep( int( arg[ '--delay' ] ) )
                    except KeyboardInterrupt :
                        pass
            elif arg[ 'log' ] :
                rea.get_log( arg[ '<chunk_id>' ], arg[ '--dir' ] )
            elif arg[ 'info' ] :
                rea.information()
            elif arg[ 'set-priority' ] :
                rea.set_priority( int( arg[ '<priority>' ] ) )
            elif arg[ 'set-restart' ] :
                rea.set_restart( arg[ '<date>' ] )
            elif arg[ 'get-restart' ] :
                rea.get_restart( )
            else :
                rea.cancel( arg[ '--hard' ] )
            if arg[ '--dry-run' ] :
                session.rollback()
            else :
                session.commit()
                if arg[ 'submit' ] :
                    if arg["--wps-only"]:
                        rea.release_wps()
                    else:
                        rea.release()
                    session.commit()
                
    except OperationalError as err :
        logging.error( err.message )
    except KeyboardInterrupt :
        session.rollback()
    except Exception as err :
        session.rollback()
        logging.error( str( err ) )
    finally:
        session.close()



#
# Copyright 2016 Universidad de Cantabria
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import os
import socket
from os.path              import exists, join, isfile

def get_hostname():
    """
    Hostname info 
    """
    return socket.gethostname()

def os_release():
    """
    Get OS release
    """
    release = ''
    for file in [ 'debian_version' , 'lsb-release' , 'redhat-release' ] :
        file_name = join( '/etc', file ) 
        if isfile( file_name ):
            f = open( file_name, 'r' )
            release = f.read().strip()
            f.close()
    return release
            
def cpu_info():
    """
    Get CPU information
    """
    model_name     = ''
    number_of_cpus = 0
    file_name = '/proc/cpuinfo'
    if exists( file_name ) :
        f = open( file_name, 'r' )
        cpu = f.readlines()
        f.close()
        number_of_cpus = 0
        for line in cpu :
            if line.startswith( 'model name' ): 
                model_name = line.rstrip('\n').split(':')[1]
            elif line.find( 'processor' ) == 0:
                number_of_cpus = number_of_cpus + 1
    return model_name, number_of_cpus

def mem_info():
    """
    Get memory information
    """
    total_mem = 0
    file_name = '/proc/meminfo' 
    if exists( file_name ):
        f = open( file_name, 'r' )
        mem = f.readlines()
        f.close()
        for line in mem :
            if line.find( 'MemTotal:' ) == 0:
                total_mem = int( line.split()[1] ) / 1024
    return total_mem 

def disk_space_check( directory ):
    """
    Disk space check
    """
    fs = os.statvfs( directory )
    disk_space = fs[4] * fs[0] / 1024 / 1024 / 1024 
    return disk_space



class MockStat(object):
    st_mode = 0


def mock_makedirs(idir):
    print("\nRunning os.makedirs({})".format(idir))


def mock_chmod(ifile, perm):
    print("\nRunning os.chmod({}, {})".format(ifile, perm))


def mock_stat(ifile):
    print("\nRunning os_stat({})".format(ifile))
    return MockStat()


def mock_chdir(idir):
    print("\nRunning os.chdir({})".format(idir))


def mock_rmtree(idir):
    print("\nRunning shutil.rmtree({})".format(idir))


def mock_copy_file(orig, dest):
    print("\nRunning copy_file with {} and {}".format(orig, dest))


def mock_shutil_copyfile(orig, dest):
    print("\nRunning shutil.copyfile with {} and {}".format(orig, dest))


def mock_extract_file(dest, to_path):
    print("\nExtracting file {} to {}".format(dest, to_path))


def mock_exec_cmd(app_cmd):
    print("\nExecuting command {}".format(app_cmd))
    if ">" in app_cmd:
        logfile = app_cmd.split(">")[1].strip()
        with open(logfile, "w") as lf:
            lf.write("Successful completion\n")
            lf.write("SUCCESS COMPLETE")
    code = 0
    output = "Successful completion"
    return code, output


import logging
import mock
import os
import shutil
import sys
import tempfile
import unittest
from wrf4g.wrapper import PilotParams, WRF4GWrapper
from unittest import mock
from mocked_functions import mock_rmtree, mock_exec_cmd

log = logging.getLogger()
log.setLevel("DEBUG")

testargs = [
    "../../bin/wrf_wrapper.py",
    "test_experiment",
    "test_realization",
    "1",
    "2001-01-01_00:00:00",
    "2001-01-07_00:00:00",
    "0"
]
os.environ["GW_HOSTNAME"] = "bender"
os.environ["GW_JOB_ID"] = "0"
os.environ["GW_RESTARTED"] = "0"
os.environ["GW_NP"] = "4"


def write_phony_rsl(tempdir):
    rsldir = tempdir + "/WRFV3/run"
    os.makedirs(rsldir)
    rslfile = rsldir + "/rsl.out.0000"
    with open(rslfile, "w") as rsl:
        rsl.write("SUCCESS COMPLETE\n")


@mock.patch.object(sys, 'argv', testargs)
def get_wrf4g_wrapper(tempdir):
    params = PilotParams(
        json="test_experiment/realization.json",
        root_path=tempdir
    )
    return WRF4GWrapper(params)


def create_tempdir(tempdir):
    log.info("Running test in {}".format(tempdir))
    bindir = tempdir + "/bin"
    os.mkdir(bindir)
    shutil.copy2("./test_experiment/namelist.input", tempdir)
    shutil.copy2("./test_experiment/wrf4g_files/bin/preprocessor.default",
                 bindir)
    shutil.copy2("./test_experiment/wrf4g_files/bin/postprocessor.SFC", bindir)
    write_phony_rsl(tempdir)


class TestWRF4GWrapper:
    #@mock.patch('wrf4g.wrapper.os.makedirs', side_effect=mock_makedirs)
    #@mock.patch('wrf4g.wrapper.os.chmod', side_effect=mock_chmod)
    #@mock.patch('wrf4g.wrapper.os_stat', side_effect=mock_stat)
    #@mock.patch('wrf4g.wrapper.os.chdir', side_effect=mock_chdir)

    #@mock.patch('wrf4g.wrapper.shutil.copyfile',
    #            side_effect=mock_shutil_copyfile)
    #@mock.patch('wrf4g.wrapper.copy_file', side_effect=mock_copy_file)
    #@mock.patch('wrf4g.wrapper.extract', side_effect=mock_extract_file)
    @mock.patch('wrf4g.wrapper.shutil.rmtree', side_effect=mock_rmtree)
    @mock.patch('wrf4g.wrapper.exec_cmd', side_effect=mock_exec_cmd)
    @mock.patch('wrf4g.wrapper.fix_ptop')
    @mock.patch('wrf4g.utils.namelist.get_latlon_dx')
    def test_launch(self, *args):
        tempdir = tempfile.mkdtemp(prefix="wrf4g_test_", dir="/tmp")
        create_tempdir(tempdir)
        wrf4g_wrapper = get_wrf4g_wrapper(tempdir)

        try:
            wrf4g_wrapper.launch()
        except SystemExit as sysexit:
            # This will always happen
            # Clean test folder
            shutil.rmtree(tempdir)


if __name__ == "__main__":
    test_wrf4g_wrapper = TestWRF4GWrapper()
    test_wrf4g_wrapper.test_launch()



#
# Copyright 2016 Universidad de Cantabria
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

from sqlalchemy     import ( Column, INTEGER,
                             VARCHAR, SMALLINT,
                             DATETIME, ForeignKey,
                             Interval, PickleType, 
                             MetaData, Table )
from sqlalchemy.orm import relationship, mapper
from wrf4g.core     import ( Experiment, Realization,
                             Chunk, Job, Events )

#
#  WRF4G database is composed of the following tables:
#  experiment
#  realization
#  chunk
#  job
#  events
#

metadata = MetaData()

experiment = Table( 'experiment', metadata,
                    Column('id',             INTEGER, primary_key=True, nullable=False),
                    Column('name',           VARCHAR(length=512), nullable=False),
                    Column('home_directory', VARCHAR(length=300))
                  )                                           

realization = Table( 'realization', metadata,
                     Column('id',               INTEGER, primary_key=True, nullable=False),
                     Column('exp_id',           INTEGER, ForeignKey('experiment.id')),
                     Column('name',             VARCHAR(length=1024), nullable=False),
                     Column('start_date',       DATETIME()),
                     Column('end_date',         DATETIME()),
                     Column('chunk_size',       Interval),
                     Column('restart',          DATETIME()),
                     Column('status',           VARCHAR(length=20)),
                     Column('wps_only_status',  VARCHAR(length=20)),
                     Column('current_date',     DATETIME()),
                     Column('current_chunk',    INTEGER),
                     Column('nchunks',          INTEGER),
                     Column('cfg',              PickleType)
                   )

chunk = Table( 'chunk', metadata,
                Column('id',         INTEGER, primary_key = True, nullable = False),
                Column('rea_id',     INTEGER, ForeignKey('realization.id')),
                Column('start_date', DATETIME()),
                Column('end_date',   DATETIME()),
                Column('wps',        INTEGER),
                Column('chunk_id',   INTEGER),
                Column('status',     VARCHAR(length=20)),
              )

job = Table( 'job', metadata,
             Column('id',           INTEGER, primary_key=True, nullable=False),
             Column('gw_job',       INTEGER),
             Column('gw_restarted', INTEGER),
             Column('chunck_id',    INTEGER, ForeignKey('chunk.id')),
             Column('resource',     VARCHAR(length=45)),
             Column('status',       VARCHAR(length=20)),
             Column('exitcode',     VARCHAR(length=20))
           )

events = Table( 'events', metadata, 
                Column('id',         INTEGER, primary_key=True, nullable=False),
                Column('job_id',     INTEGER, ForeignKey('job.id')),
                Column('job_status', VARCHAR(length=20)),
                Column('timestamp',  DATETIME())
              )

mapper(Experiment, experiment, properties={
       "realization" : relationship(Realization, back_populates="experiment", lazy='dynamic')
})
mapper(Realization, realization, properties={
       "experiment" : relationship(Experiment, back_populates = "realization" ),
       "chunk"      : relationship(Chunk,      back_populates = "realization", lazy='dynamic')
})
mapper(Chunk, chunk, properties={
       "realization" : relationship(Realization, back_populates = "chunk"),
       "job"         : relationship(Job,         back_populates = "chunk", lazy='dynamic')
       
})
mapper(Job, job, properties={
       "chunk"  : relationship(Chunk,  back_populates = "job"),
       "events" : relationship(Events, back_populates = "job", lazy='dynamic')
})
mapper(Events, events, properties={
       "job" : relationship(Job, back_populates = "events")
})





#
# Copyright 2016 Universidad de Cantabria
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import os 
import datetime
import logging
from re                  import match, search
from os.path             import abspath, isdir, isfile, basename, dirname, join
from wrf4g.utils.command import exec_cmd

def http_ftp_protocol(protocol):
    if protocol == 'http' or protocol == 'https' or protocol == 'ftp':
        return True
    else :
        return False

class VCPURL(object):
    """
    Examples of usage of the VCPURL class:
    
    GSIFTP:
    >>> c=VCPURL("gsiftp://ce01.macc.unican.es:2812/tmp/example")
    >>> print c
    gsiftp://ce01.macc.unican.es:2812/tmp/example
    >>> c.rename("example1")
    >>> a=VCPURL("gsiftp://ce01.macc.unican.es:2812/tmp")
    >>> a.ls("exam*")
    ['example1']
    
    FILES:
    >>> d=VCPURL("./new_directory")
    >>> d.mkdir()
    >>> d.ls("*")
    ['']
    >>> f=VCPURL(".")
    >>> f.ls("new*")
    ['new_directory']
    
    RSYNC:
    >>> a=VCPURL("rsync://user@ce01.macc.unican.es/tmp/")
    >>> a.ls("expam*")
    ['example2']
    >>> a.rename("example2p")
    >>> a.ls("example*")
    ['example2p']
    """
    
    def __init__(self, url=None):
        """
        From a url (protocol://user@machine:port/file) it returns an array with 
        the following field [protocol,user,computer,port,file]
        """
        self.protocol     = ""
        self.user         = ""
        self.computer     = ""
        self.port         = ""
        self.file         = ""
        self.command      = ""
        self.usercomputer = ""
        
        if not url:
            out = "Not an url"
            raise Exception(out)
        else:
            # Split the url if the begging is protocol:
            g0 = match("([a-z]+):(.+)", url)
            # There's no protocol in the url
            if not g0 :
                self.protocol = "file"
                abs_url = abspath(url)
                if url == ".":
                    self.file = abs_url + "/"
                else:
                    self.file = abs_url
            else :
                # The url contains a protocol
                (self.protocol, self.file) = g0.groups()
                # This is from URL with the form: gsiftp://ce01.macc.unican.es
                position = self.file.find('//')
                if position == 0 :
                    g1 = match("//([\w.@-]*):?(\d*)(/\S*)", self.file)
                    if g1 :
                        (self.usercomputer, self.port, self.file) = g1.groups()
                        self.computer = self.usercomputer
                        if self.protocol in ("ln", "file") :
                            self.file = abspath (self.file)
                        if self.computer.find("@") != -1 :
                            (self.user, self.computer) = self.computer.split("@")
                    else :
                        out="Url is not well formed"
                        raise Exception(out)
                    
            self.command = {'file'  : {'ls'    : "'ls -1 %s'    %self.file" ,
                                       'mkdir' : "'mkdir -p %s' %self.file" ,
                                       'rm'    : "'rm -rf %s'   %self.file" , 
                                       'rename': "'mv %s %s'    %(orig,dest)", 
                                       'name'  : "self.file"},
                            'http'  : {'ls'    : "" ,
                                       'mkdir' : "" ,
                                       'rm'    : "" , 
                                       'rename': "" , 
                                       'name'  : "self.file"},
                            'https'  : {'ls'   : "" ,
                                       'mkdir' : "" ,
                                       'rm'    : "" , 
                                       'rename': "" , 
                                       'name'  : "self.file"},
                            'ftp'   : {'ls'    : "" ,
                                       'mkdir' : "" ,
                                       'rm'    : "" , 
                                       'rename': "" , 
                                       'name'  : "self.file"},   
                            'sftp'  : {'ls'    : '"echo \'ls %s\'        | sftp %s" %(self.file, self.usercomputer)' ,
                                       'mkdir' : '"echo \'mkdir %s\'     | sftp %s" %(self.file, self.usercomputer)' ,
                                       'rm'    : '"echo \'rm %s\'        | sftp %s" %(self.file, self.usercomputer)' ,
                                       'rename': '"echo \'rename %s %s\' | sftp %s" %(orig,dest, self.usercomputer)' ,
                                       'name'  : "self.file"}, 
                            'rsync' : {'ls'    : "'ssh -q %s ls -1 %s'    %(self.usercomputer,self.file)",
                                       'mkdir' : "'ssh -q %s mkdir -p %s' %(self.usercomputer,self.file)", 
                                       'rm'    : "'ssh -q %s rm -rf %s'   %(self.usercomputer,self.file)", 
                                       'rename': "'ssh -q %s mv %s %s'    %(self.usercomputer,orig,dest)", 
                                       'name'  : "self.file"},
                            'gsiftp': {'ls'    : "'uberftp %s \"ls %s\" '         %(self.computer,self.file)", 
                                       'mkdir' : "'uberftp %s \"mkdir %s\"'       %(self.computer,self.file)", 
                                       'rm'    : "'uberftp %s \"rm -r %s\"'       %(self.computer,self.file)", 
                                       'rename': "'uberftp %s \"rename %s %s\"'   %(self.computer,orig,dest)",
                                       'name'  : "self.file"},
                            'lfn'   : {'ls'    : "'lfc-ls %s'         %(self.file)",
                                       'mkdir' : "'lfc-mkdir -p %s'   %(self.file)",
                                       'rm'    : "'lcg-del -a lfn:%s' %(self.file)",
                                       'rename': "'lfc-rename %s %s'  %(orig,dest)",
                                       'name'  : "self.file"}
                            }
            
    def __str__(self):
        """
        Returns an url from a array with the following fields 
        [protocol,user,computer,port,file]
        """
        # Fill the variables: protocol, user, computer, port and file 
        # user
        if self.user != "" : 
            user = self.user + "@"
        else : 
            user = ""
        
        # computer 
        computer = self.computer
        
        # port 
        if self.port != "" or self.protocol == "rsync" : 
            port = ":" + self.port
        else : 
            port = ""
            
        # file 
        file = self.file
        
        # protocol
        if self.protocol == "rsync" :
            protocol = ""
        # Transform local path to file://. If the file is a directory we add / at the end because
        # globus-url-copy works this way.
        elif self.protocol == "file":
            protocol = self.protocol + "://"
            file = abspath(file)
        else : protocol = self.protocol + "://"
        
        url = protocol + user + computer + port + file 
        return url
    
    def ls(self, file="*"):
        """ 
        List all the files under the directory. If no arguments are given it list the whole directory. 
        It has the argument file that is the pattern we want to search inside this directory. Examples
            * a.ls()
            * a.ls("file*")
        """
        if http_ftp_protocol(self.protocol):
            out="This method is not available for " + self.protocol + " protocol"
            raise Exception(out)
        
        command = eval(self.command[self.protocol]['ls'])
        (err, out) = exec_cmd(command)
        if err :
            out = "Error reading dir: " + str(out)
            raise Exception(out)
        
        if self.protocol == 'sftp' and 'not found' in out :
            return []
        out_list = out.split("\n")
        if self.protocol == "gsiftp":
            aux_list = []
            for i, elem in enumerate(out_list[2:]):
                aux_list.append( elem.split()[-1] )
            out_list = aux_list

        pattern = basename( file ).replace(".", "\.")
        pattern = pattern.replace("*", ".*")
        pattern = pattern + "$"

        try :
            out_list.remove(".")
            out_list.remove("..")
        except ValueError:
            pass

        file_list = []
        for file in out_list :
            file_name = basename( file )
            if match(pattern, file_name) :
                file_list.append(file_name)
        file_list.sort()
        return file_list
    
    def mkdir(self, mode=None):
        """
        Create the directory pointed by self with mode permissions if required

        """
        if http_ftp_protocol(self.protocol):
            out="This method is not available for " + self.protocol + " protocol"
            raise Exception(out)
        
        command = eval(self.command[self.protocol]['mkdir'])
        if mode is not None:
            command = command.replace("mkdir", "mkdir -m {}".format(mode))
        (err, out) = exec_cmd(command)
        if err :
            out = "Error creating dir: " + str(out)
            raise Exception(out)
    
    def rm(self):
        """
        Delete a file or folder
        """
        if http_ftp_protocol(self.protocol):
            raise Exception("This method is not available for " + self.protocol + " protocol")
        
        command = eval(self.command[self.protocol]['rm'])
        (err, out) = exec_cmd(command)
        if err : 
            out="Error deleting file: " + str(out)
            raise Exception(out)
    
    def rename(self, newname):
        """
        Rename self.file into newname
        """
        if http_ftp_protocol(self.protocol):
            out="This method is not available for " + self.protocol + " protocol"
            raise Exception(out)
        
        orig = eval(self.command[self.protocol]['name'])
        dest_folder = dirname(orig)
        dest = join(dest_folder,newname)
        command = eval(self.command[self.protocol]['rename'])
        (err, out) = exec_cmd(command)
        if err :
            out="Error listing file: " + str(out)
            raise Exception(out)

    def isfile(self):
        """
        Return True if self.file is a file
        """
        if self.protocol == "ln" :
           return True
        elif self.protocol == "sftp" and '*' in self.file :
           return True
        elif self.protocol == "file" :
           if isfile( self.file ) :
               return True
           else :
               return False
        else:
           out = self.ls(self.file)
           if len(out) != 0 and basename(self.file) == out[0]:
              return True
           else:
              return False
    
    def exists(self):
        """
        Return True if self.file exists
        """
        try:
            if not self.ls(self.file):
                return False
        except Exception :
            return False
        else:
            return True

def copy(origin, destination, overwrite=True):
    """
    Copy orig to dest.
    """
    if VCPURL(origin).isfile():
        copy_file(origin, destination, overwrite)
    else:
        copy_tree(origin, destination, overwrite)


def copy_tree(origin, destination, overwrite=True):
    """
    Recursively copy orig to dest.
    """
    orig=VCPURL(origin)
    dest=VCPURL(destination)
    if dest.exists():
        dest.mkdir()
    for name in orig.ls():
        if not name: break
        orig_name = join(orig.__str__(), name)
        dest_name = join(dest.__str__(), name)
        try:
            if VCPURL(orig_name).isfile():
                copy_file(orig_name, dest_name, overwrite)
            else:
                copy_tree(orig_name, dest_name, overwrite)
        except Exception as err:
            logging.warning(str(err))

def copy_file(origin, destination, overwrite=True):
    """
    Copy orig to dest file.
    """
    vcp_matrix = {'file': {'file':   {'verbose'  : '-v', 
                                  'command'  : "'cp -f -p %(verbose)s %(orig)s %(dest)s' %param",
                                  'orig'     : "orig.file", 
                                  'dest'     : "dest.file"},
                       'rsync':  {'verbose'  : '-v -p',
                                  'command'  : "'rsync -au %(verbose)s %(orig)s %(dest)s' %param", 
                                  'orig'     : "orig.file", 
                                  'dest'     : "str(dest)"},
                       'ln':     {'verbose'  : '-v', 
                                  'command'  : "'ln -s %(verbose)s %(orig)s %(dest)s' %param", 
                                  'orig'     : "orig.file",
                                  'dest'     : "dest.file"},
                       'gsiftp': {'verbose'  : '-v', 
                                  'command'  : "'globus-url-copy %(verbose)s %(orig)s %(dest)s' %param", 
                                  'orig'     : "str(orig)", 
                                  'dest'     : "str(dest)"},
                       'sftp':   {'verbose'  : '',
                                  'command'  : '"echo \'put -P %(orig)s %(dest)s\' | sftp %(usercomputer)s" %param',
                                  'orig'     : "orig.file",
                                  'usercomputer' : "dest.usercomputer",
                                  'dest'     : "dest.file"},
                       'lfn':    {'verbose'  : '-v',
                                  'command'  : "'lcg-cr %(verbose)s %(dest)s %(orig)s' %param",
                                  'orig'     : "str(orig)",
                                  'dest'     : "str(dest)"},
                       },
              'gsiftp':{'file':  {'verbose'  : '-v', 
                                  'command'  : "'globus-url-copy %(verbose)s %(orig)s %(dest)s' %param", 
                                  'orig'     : "str(orig)", 
                                  'dest'     : "str(dest)"},
                        },
              'rsync': {'file' : {'verbose'  : '-v', 
                                  'command'  : "'rsync -au %(verbose)s %(orig)s %(dest)s' %param", 
                                  'orig'     : "str(orig)", 
                                  'dest'     : "dest.file"},
                        },
              'sftp':  {'file' : {'verbose'  : '',
                                  'command'  : '"echo \'get -P %(orig)s %(dest)s\' | sftp %(usercomputer)s" %param',
                                  'orig'     : "orig.file",
                                  'usercomputer' : "orig.usercomputer",
                                  'dest'     : "dest.file"},
                        },
              'https': {'file' : {'verbose'  : '-v', 
                                  'command'  : "'wget %(verbose)s %(orig)s %(dest)s' %param", 
                                  'orig'     : "str(orig)", 
                                  'dest'     : "dest.file"},
                        },
              'http': {'file' : {'verbose'   : '-v', 
                                  'command'  : "'wget %(verbose)s %(orig)s %(dest)s' %param", 
                                  'orig'     : "str(orig)", 
                                  'dest'     : "dest.file"},
                       },
              'ftp' : {'file' : {'verbose'   : '-v', 
                                  'command'  : "'wget %(verbose)s %(orig)s %(dest)s' %param", 
                                  'orig'     : "str(orig)", 
                                  'dest'     : "dest.file"},
                       },
              'lfn' : {'file' : {'verbose'   : '-v',
                                  'command'  : "'lcg-cp %(verbose)s lfn:%(orig)s %(dest)s' %param",
                                  'orig'     : "orig.file",
                                  'dest'     : "str(dest)"},
                       },
              }

    orig = VCPURL(origin)
    #For protocols like lfn the destination must be a file
    if destination[-1] == "/" :
        dest = VCPURL(join(destination, basename(orig.file)))
    elif  destination == ".":
        dest = VCPURL(basename(orig.file))
    else:
        dest = VCPURL(destination)
    out="Starting to copy ..."
    logging.debug(out)
    if http_ftp_protocol(dest.protocol):
        out="Unable to copy if the destination protocol is " + dest.protocol
        raise Exception(out)
    if http_ftp_protocol(orig.protocol) and dest.protocol != 'file':
        out="Unable to copy if the destination protocol is not file://"
        raise Exception(out)
    if dest.protocol == 'ln' and orig.protocol != 'file':
        dest.protocol = 'file'

    #If "orig" and "dest" are on the same machine, VCP will copy using cp command
    if (dest.protocol == "rsync" or orig.protocol == "rsync") :
        if (isfile(orig.file) and (isdir(dest.file) or isdir(dest.file.rsplit('/',1)[0]))) or \
           (isdir(orig.file)  and isdir(dest.file)) or \
           ("*" in orig.file  and isdir(orig.file.rsplit('/',1)[0]) and isdir(dest.file)):
            orig.protocol = "file"
            dest.protocol = "file"
    out = "Copying from " + orig.__str__() + " to " + dest.__str__()
    logging.debug(out)   
    param = vcp_matrix[orig.protocol][dest.protocol]
    #orig customization 
    orig_file = eval(param['orig'])
    param['orig'] = orig_file
    #dest customization 
    if dest.protocol == "lfn" and dest.computer:
        dest_file = '-d ' + dest.computer + ' -l ' + dest.file
    else:
        dest_file = eval(param['dest'])
    param['dest'] = dest_file
    if orig.protocol == "sftp" or dest.protocol == "sftp" :
       param['usercomputer'] = eval(param['usercomputer'])
    if logging.DEBUG : 
        param['verbose'] = ""
    #Overwrite the destination if necessary
    if overwrite:
        if dest.exists() and dest.isfile():
            out = "Overwriting " + dest.__str__()
            logging.debug(out)
            dest.rm()
    start = datetime.datetime.now()
    command = eval(param['command'])
    out="Command to copy... " + command
    (err, out) = exec_cmd(command)
    if err :
        raise Exception("Error copying file: " + out)
    end = datetime.datetime.now()
    out = "The copy lasted %s" % ( end - start )
    logging.debug(out)




#
# Copyright 2016 Universidad de Cantabria
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

"""
Print information about the hosts available on WRF4G.
     
Usage: 
    wrf4g host [ --dbg ] [ list ] [ <hid> ] 
    
Arguments:
    <hid>         Host identifier.

Options:
    --dbg         Debug mode.        
 
Host field information:
    HID           Host identifier.
    ARCH          Architecture.
    JOBS(R/T)     Number of jobs: R = running, T = total.
    LRMS          Local Resource Management System.
    HOSTNAME      Host name. 
    QUEUENAME     Queue name.
    WALLT         Queue wall time.
    CPUT          Queue cpu time.
    MAXR          Max. running jobs.
    MAXQ          Max. queued jobs. 
"""

import logging
import sys
from drm4g.commands       import exec_cmd, Daemon

def run( arg ) :
    logging.basicConfig( format = '%(message)s', 
                         level  = logging.DEBUG if arg[ '--dbg' ] else logging.INFO,
                         stream = sys.stdout )
    try :
        daemon = Daemon()
        if not daemon.is_alive() :
            raise Exception('DRM4G is stopped.')
        cmd = 'gwhost '
        if arg[ '<hid>' ] :
            cmd = cmd + arg[ '<hid>' ]
        out , err = exec_cmd( cmd )
        logging.info( out )
        if err :
            logging.info( err )
    except KeyboardInterrupt :
        pass
    except Exception as err :
        logging.error( str( err ) )




from drm4g.managers.slurm import *

class Job(drm4g.managers.slurm.Job):
    def jobTemplate(self, parameters):
        args  = '#!/bin/bash\n'
        args += '#SBATCH --job-name=JID_%s\n' % (parameters['environment']['GW_JOB_ID'])
        args += '#SBATCH --output=$stdout\n'
        args += '#SBATCH --error=$stderr\n'
        if parameters['queue'] != 'default':
            args += '#SBATCH -p $queue\n'
        if 'maxWallTime' in parameters :
            args += '#SBATCH --time=%s\n' % (parameters['maxWallTime'])
        if 'maxMemory' in parameters :
            args += '#SBATCH --mem=%s\n' % (parameters['maxMemory'])
        if 'ppn' in parameters :
            args += '#SBATCH --ntasks-per-node=$ppn\n'
        if 'nodes' in parameters :
            args += '#SBATCH --nodes=$nodes\n'
        args += '#SBATCH --ntasks=$count\n'
        args += '#SBATCH --mail-type=begin\n'
        args += '#SBATCH --mail-type=end\n'
        args += '#SBATCH --mail-user=milovacj@unican.es\n'
        args += ''.join(['export %s=%s\n' % (k, v) for k, v in list(parameters['environment'].items())])
        args += '\n'
        args += '$executable\n'
        return Template(args).safe_substitute(parameters)



#
# Copyright 2016 Universidad de Cantabria
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import re
import calendar
from datetime                import datetime, timedelta
from dateutil.relativedelta  import relativedelta

def datewrf2datetime( date_wrf ):
    mo = re.match("(\d{4})-(\d{2})-(\d{2})_(\d{2})[:_](\d{2})[:_](\d{2})", date_wrf )
    if not mo :
        raise Exception("Date is not well formed")
    date_tuple  = mo.groups()
    date_object = datetime( *tuple( map( int, date_tuple ) ) )
    return date_object

def dateiso2datetime( date_iso ):
    mo = re.match("(\d{4})(\d{2})(\d{2})T(\d{2})(\d{2})(\d{2})Z", date_iso )
    if not mo :
        raise Exception("Date is not well formed")
    date_tuple  = mo.groups()
    date_object = datetime( *tuple( map( int, date_tuple ) ) )
    return date_object

def datetime2datewrf( date_object ):
    return date_object.strftime("%Y-%m-%d_%H:%M:%S")

def datetime2dateiso( date_object ):
    return date_object.strftime("%Y%m%dT%H%M%SZ")

def str2timedelta( input_str ):
    keys = [ "years", "months", "days", "hours" ]
    match = False
    for key in keys :
        if key in input_str :
            match = True
    if not match :
        raise Exception( "'%s' does not contain an argument. Please add an argument may be years, months, days, or hours." % input_str ) 
    regex = "".join(["((?P<%s>\d+)\s*%s ?)?" % (k, k[0]) for k in keys])
    kwargs = {}
    for key, val in re.match( regex, input_str ).groupdict( default = "0" ).items():
        kwargs[ key ] = int( val )
    return relativedelta( **kwargs )

def timedelta_total_seconds( td ):
    try :
        return (td.microseconds + (td.seconds + td.hours * 3600 + td.days * 24 * 3600) * 10**6) / 10**6     
    except :
        return (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) / 10**6

class Calendar( object ):
    """
    Class to manage calendars like 'standard' and 'no_leap' 
    """
    available_types = ( 'standard', 'no_leap' )

    def __init__( self, type = 'standard' ):
        if not type in self.available_types :
            raise Exception( "Calendar type not available" )
        else :
            self.type = type

    def sub_dates(self, date1, date2 ):
        """
        Subtract two dates returning a timedelta object
        """
        if self.type == 'no_leap' :
            diff_days = self._no_leap_day( date1 ) - self._no_leap_day( date2 )
            return relativedelta( days=diff_days )
        else :
            return date1 - date2

    def add(self, date, trelativedelta):
        """
        Add time to a date returning a datetime object
        """
        total_time = date + trelativedelta
        if self.type == 'no_leap' :
            hours_to_add = 0
            for year in range( date.year, total_time.year + 1 ) :
                if calendar.isleap( year ) : 
                    if date < datetime( year, 2, 29 ) < total_time :
                        hours_to_add += 24
            total_time += relativedelta( hours = hours_to_add ) 
        return total_time

    def sub(self, date, trelativedelta):
        """
        Subtract time to a date returning a datetime object
        """
        total_time = date - trelativedelta
        if self.type == 'no_leap' :
            hours_to_sub = 0
            for year in range( date.year, total_time.year + 1 ) :
                if calendar.isleap( year ) :     
                    if date < datetime( year, 2, 29 ) < total_time :
                        hours_to_sub -= 24
            total_time -= relativedelta( hours = hours_to_sub )
        return total_time

    def _no_leap_day(self, date):
        year        = date.year
        month       = date.month
        day         = date.day
        hour        = date.hour
        minute      = date.minute
        second      = date.second
        microsecond = date.microsecond
        # Convert time to fractions of a day
        day = day + hour / 24.0 + minute / 1440.0 + (second + microsecond/1.e6) / 86400.0
        # Start Meeus algorithm 
        if month < 3 :
            month = month + 12
            year = year - 1
        day = int(365.0 * (year + 4716)) + int(30.6001 * (month + 1)) + day - 1524.5
        return day




#
# Copyright 2016 Universidad de Cantabria
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

"""
Manage WRF4G experiments. 
    
Usage: 
    wrf4g exp list                [ --pattern=<name> ]
    wrf4g exp <name> define       [ --dbg ] [ --force ]   [ --from-template=<name> ] [ --dir=<directory> ] 
    wrf4g exp <name> edit         [ --dbg ] 
    wrf4g exp <name> create       [ --dbg ] [ --dry-run ] [ --dir=<directory> ]
    wrf4g exp <name> update       [ --dbg ] [ --dry-run ] 
    wrf4g exp <name> submit       [ --dbg ] [ --dry-run ] [ --priority=<value> ] [ --pattern=<name> ] [ --rea-state=<state> ] [ --rerun ]  [ --wps-only ] [ --mode=<value> ] 
    wrf4g exp <name> status       [ --dbg ] [ --pattern=<name> ] [ --rea-state=<state> ] [ --delay=<seconds> ] [ --show-chunks ]
    wrf4g exp <name> statistics   [ --pattern=<name> ]
    wrf4g exp <name> cancel       [ --dbg ] [ --dry-run ] [ --pattern=<name> ] [ --rea-state=<state> ] [ --hard ]
    wrf4g exp <name> set-priority [ --dbg ] [ --dry-run ] [ --pattern=<name> ] <priority>
    wrf4g exp <name> delete       [ --dbg ] [ --dry-run ] 
   
Options:
    --dbg                     Debug mode.
    -n --dry-run              Dry run.
    -f --force                Force to remove if it exists.
    -P --priority=<value>     Fix-priority for scheduling [default: 0]. 
    -p --pattern=<name>       Pattern to find experiments and realizations.
    -s --rea-state=<state>    Select only realizations in the indicated state. Available states :
                              PREPARED, SUBMITTED, RUNNING, PENDING, FAILED and FINISHED 
    -t --from-template=<name> Experiment template, avaible templates are default, single, physics. 
    -d --dir=<directory>      Directory to create or start an experiment [default: ./].
    --delay=<seconds>         Refresh experiment information every delay seconds.
    --rerun                   Force to run although this realization or experiment has finished.
    --hard                    Remove jobs from without synchronizing.
    -m --mode=<value>         0: run whole workflow, 1: only WPS and real, 2: only WRF [default: 0]
    --show-chunks                  Show advanced information about the chunks status
  
Commands:
    list                      Show all the experiments available.
    define                    Create the files needed to define a WRF4G experiment.
    edit                      Edit experiment.wrf4g file.
    create                    Given experiment.wrf4g file, prepare the experiment creating 
                              the realizations and chunks needed.
    update                    Update the experiment configuration.
    submit                    Submit the experiment.
    status                    Check the status of realizations and chunks showing computing resources, 
                              job identifier and exit codes (SEE EXIT CODES) 
    cancel                    Cancel the active realizations by killing their jobs.
    set-priority              Change the scheduling priority of any job releted to a realization. 
                              The priority must be in range [0,20], and the default value is 0. 
                              When a job gets a priority of 20, it becomes an urgent job. This job 
                              is dispatched as soon as possible passing all the scheduling policies.
    delete                    Remove the experiment from the database.

EXIT CODES
    1  : Error creating log directory        
    2  : Error copying apps            
    3  : Error app type does not exist            
    4  : Error executing source script       
    5  : Job already executed 
    6  : Error creating directory to simulate
    7  : Error finding WRF binaries 
    8  : Error copying restart files        
    9  : There is a mismatch in the restart date   
    10 : Error copying files    
    11 : Error downloading WPS files    
    12 : Error copying boundaries           
    13 : Error modifying namelist
    14 : Error executing PREPROCESSOR
    15 : Error linking GRIB files     
    16 : Error executing UNGRIB
    17 : Error executing METGRID       
    18 : Error executing REAL
    19 : Error uploading WPS files      
    20 : Error executing WRF
    21 : Error executing POSTPROCESSOR 
    22 : Error copying output file    
    23 : Job killed by the system 
    255: Unexpected error
"""

import sys
import time
import logging
import shutil
from sqlalchemy.orm.exc   import NoResultFound
from sqlalchemy.exc       import OperationalError
from wrf4g.db             import get_session
from wrf4g.core           import Experiment
from wrf4g.utils.command  import cls
from wrf4g.utils.time     import datetime2datewrf

def run( arg ) :
    logging.basicConfig( format = '%(message)s', 
                         level  = logging.DEBUG if arg[ '--dbg' ] else logging.DEBUG,  
                         stream = sys.stdout )
    if arg[ 'define' ] :
        Experiment.create_files( arg[ '<name>' ], 
                                 "default" if not arg[ '--from-template' ] else arg[ '--from-template' ], 
                                 arg[ '--force' ], arg[ '--dir' ] )
    else :
        # create a session to connect with the database
        session = get_session()
        try :
            if arg[ 'list' ] :
                l_exps = session.query( Experiment )
                if arg[ '--pattern' ] :
                    l_exps  = l_exps.filter( Experiment.name.like( arg[ '--pattern' ].replace('*','%') ) )
                if not l_exps.all() :
                    logging.info( "There are not experiments" )
                else :
                    # Header
                    logging.info( "\033[1;4m%-40s\033[0m" % ( "EXPERIMENT" ) )
                    for e in l_exps :
                        logging.info("%-s" % ( e.name ) ) 
            else :
                try :
                    exp = session.query( Experiment ).\
                            filter_by( name = arg[ '<name>' ] ).one()
                    exp.dryrun = arg[ '--dry-run' ]
                except NoResultFound :
                    if arg[ 'create' ] :
                        exp2        = Experiment()
                        exp2.name   = arg[ '<name>' ]
                        exp2.dryrun = arg[ '--dry-run' ]
                        exp2.prepare( update = False, directory = arg[ '--dir' ] )
                        session.add( exp2 )
                    else :
                        raise Exception( "'%s' experiment does not exist" % arg[ '<name>' ] )
                else :
                    # Options 
                    if arg[ 'edit' ] :
                        exp.edit( )
                    if arg[ 'update' ] :
                        exp.prepare( update = True )
                    elif arg[ 'submit' ] :
                        if arg['--wps-only']:
                            exp.run_wps( 
                                rerun           = arg[ '--rerun' ],
                                priority        = int( arg[ '--priority' ] ), 
                                )
                        else:
                            exp.run( arg[ '--rerun' ], 
                                 arg[ '--pattern' ], 
                                 arg[ '--rea-state' ],
                                 int( arg[ '--priority' ] )
                                 )
                    elif arg[ 'status' ] :
                        if not arg[ '--delay' ] :
                            exp.get_status( arg[ '--pattern' ], arg[ '--rea-state' ],arg['--show-chunks'] )
                        else :
                            try:
                                while True :
                                    cls()
                                    exp.get_status( arg[ '--pattern' ], arg[ '--rea-state' ] )
                                    time.sleep( int( arg[ '--delay' ] ) )
                            except KeyboardInterrupt : 
                                pass
                    elif arg[ 'cancel' ] :
                        exp.cancel( arg[ '--pattern' ], arg[ '--rea-state' ], arg[ '--hard' ] )
                    elif arg[ 'set-priority' ] :
                        exp.set_priority( arg[ '--pattern' ], int( arg[ '<priority>' ] ) )
                    elif arg[ 'statistics' ] :
                        exp.statistics( arg[ '--pattern' ] )
                    elif arg[ 'delete' ] :
                        exp.delete( )
                        session.delete( exp )
                        logging.info( "'%s' experiment has been deleted from the database" % exp.name )
                    elif arg[ 'create' ] :
                        logging.info( "'%s' experiment already exists" % exp.name )
                        
            if arg[ '--dry-run' ] :
                session.rollback()
            else :
                session.commit()
                if arg[ 'submit' ] :
                    if arg["--wps-only"]:
                        exp.release_wps()
                    else:
                        exp.release()
                    session.commit()
        except OperationalError as err :
            logging.error( err.message )
        except KeyboardInterrupt :
            session.rollback()
        except Exception as err :
            session.rollback()
            #logging.error( str( err ) )
            session.close()
            exit(str(err))
        finally:
            session.close()



#
# Copyright 2016 Universidad de Cantabria
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

"""
Manage identities for resources configuring private/public keys and grid credentials.

Usage:  
    wrf4g id <resource> init   [ --dbg ] [ --lifetime=<hours> ]
    wrf4g id <resource> info   [ --dbg ] 
    wrf4g id <resource> delete [ --dbg ] 

 Options:
    -l --lifetime=<hours>   Duration of the identity's lifetime [default: 168].
    --dbg                   Debug mode.
    
Commands:
    init                    Create an identity for a while, by default 168 hours 
                            (1 week). Use the option --lifetime to modify this 
                            value. It adds the configured private key to a ssh-agent
                            and creates a grid proxy using myproxy server.
                            Append the public key to the remote user's 
                            ~/.ssh/authorized_keys file (creating the file, and 
                            directory, if necessary). It tries to load the public 
                            key obtained by appending *.pub to the name of the 
                            configured private key file. Alternative the public 
                            key can be given by public_key variable.
                            It also configures the user's grid certificate 
                            under ~/.globus directory (creating directory, 
                            if necessary) if grid_cert variable is defined.
                                  
    info                    It gives some information about the identity status.
                                
    delete                  The identity is removed from the ssh-agent and the 
                            myproxy server.
"""

import logging, sys
from os.path              import expanduser, exists, expandvars
from drm4g.core.configure import Configuration
from drm4g.commands       import Daemon, Agent, Proxy

def run( arg ) :
    logging.basicConfig( format = '%(message)s',
                         level  = logging.DEBUG if arg[ '--dbg' ] else logging.INFO,
                         stream = sys.stdout )
    try :
        daemon = Daemon()
        if not daemon.is_alive() :
           raise Exception( 'DRM4G is stopped.' )
        config = Configuration()
        config.load( )
        if config.check( ) :
            raise Exception( "Review the configuration of '%s'." % ( arg['<resource>'] ) )
        if arg['<resource>'] not in config.resources :
            raise Exception( "'%s' is not a configured resource." % ( arg['<resource>'] ) )
        lrms         = config.resources.get( arg['<resource>'] )[ 'lrms' ]
        communicator = config.resources.get( arg['<resource>'] )[ 'communicator' ]
        if lrms != 'cream' and communicator != 'ssh' :
            raise Exception( "'%s' does not have an identity to configure." % ( arg['<resource>'] ) )
        if lrms == 'cream' :
            proxy = Proxy( config.resources[ arg['<resource>'] ] , 
                           config.make_communicators()[ arg['<resource>'] ] 
                           )
        if communicator == 'ssh' :
            agent = Agent( config.resources[ arg['<resource>'] ] )
        if arg[ 'init' ] :
            if communicator == 'ssh' :
                agent.start( )
                agent.add_key( arg[ '--lifetime' ] )
                agent.copy_key( )
            if lrms == 'cream' :
                proxy.configure( )
                proxy.create( arg[ '--lifetime' ] )
        elif arg[ 'delete' ] :
            if lrms == 'cream' :
                proxy.destroy( )
            if communicator == 'ssh' :
                agent.delete_key( )
        else :
            if communicator == 'ssh' :
                agent.list_key( )
            if lrms == 'cream' :
                proxy.check( )
    except KeyboardInterrupt :
        pass   
    except Exception as err :
        logging.error( str( err ) )



from datetime import datetime
from wrf4g.core import Realization


def test_realization():
    rea = Realization()
    rea_id = 15
    chunk_start_date = datetime(2011, 8, 28, 12)
    chunk_end_date = datetime(2011, 8, 30, 0)
    chunk_id = 1
    ch = rea.check_db(rea_id, chunk_start_date, chunk_end_date, chunk_id)
    print(ch)


if __name__ == "__main__":
    test_realization()